# -*- coding: utf-8 -*-
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tempfile
import unittest
import sys

import numpy as np

from pyspark.util import is_remote_only
from pyspark.ml.param import Param, Params
from pyspark.ml.tuning import ParamGridBuilder
from pyspark.sql.functions import rand
from pyspark.testing.connectutils import should_test_connect, connect_requirement_message
from pyspark.testing.utils import (
    have_sklearn,
    sklearn_requirement_message,
    have_torch,
    torch_requirement_message,
    have_torcheval,
    torcheval_requirement_message,
)
from pyspark.testing.sqlutils import ReusedSQLTestCase

if should_test_connect:
    import pandas as pd
    from pyspark.ml.connect import Model, Estimator
    from pyspark.ml.connect.feature import StandardScaler
    from pyspark.ml.connect.classification import LogisticRegression as LORV2
    from pyspark.ml.connect.pipeline import Pipeline
    from pyspark.ml.connect.tuning import CrossValidator, CrossValidatorModel
    from pyspark.ml.connect.evaluation import BinaryClassificationEvaluator, RegressionEvaluator

    class HasInducedError(Params):
        def __init__(self):
            super().__init__()
            self.inducedError = Param(
                self, "inducedError", "Uniformly-distributed error added to feature"
            )

        def getInducedError(self):
            return self.getOrDefault(self.inducedError)

    class InducedErrorModel(Model, HasInducedError):
        def __init__(self):
            super().__init__()

        def _transform(self, dataset):
            return dataset.withColumn(
                "prediction", dataset.feature + (rand(0) * self.getInducedError())
            )

    class InducedErrorEstimator(Estimator, HasInducedError):
        def __init__(self, inducedError=1.0):
            super().__init__()
            self._set(inducedError=inducedError)

        def _fit(self, dataset):
            model = InducedErrorModel()
            self._copyValues(model)
            return model


class CrossValidatorTestsMixin:
    def test_gen_avg_and_std_metrics(self):
        metrics_all = [
            [1.0, 3.0, 2.0, 4.0],
            [3.0, 2.0, 2.0, 4.0],
            [3.0, 2.5, 2.1, 8.0],
        ]
        avg_metrics, std_metrics = CrossValidator._gen_avg_and_std_metrics(metrics_all)
        assert np.allclose(avg_metrics, [2.33333333, 2.5, 2.03333333, 5.33333333])
        assert np.allclose(std_metrics, [0.94280904, 0.40824829, 0.04714045, 1.88561808])
        assert isinstance(avg_metrics, list)
        assert isinstance(std_metrics, list)

    def test_copy(self):
        dataset = self.spark.createDataFrame(
            [(10, 10.0), (50, 50.0), (100, 100.0), (500, 500.0)] * 10, ["feature", "label"]
        )

        iee = InducedErrorEstimator()
        evaluator = RegressionEvaluator(metricName="rmse")

        grid = ParamGridBuilder().addGrid(iee.inducedError, [100.0, 0.0, 10000.0]).build()
        cv = CrossValidator(
            estimator=iee,
            estimatorParamMaps=grid,
            evaluator=evaluator,
            numFolds=2,
        )
        cvCopied = cv.copy()
        for param in [
            lambda x: x.getEstimator().uid,
            # SPARK-32092: CrossValidator.copy() needs to copy all existing params
            lambda x: x.getNumFolds(),
            lambda x: x.getFoldCol(),
            lambda x: x.getParallelism(),
            lambda x: x.getSeed(),
        ]:
            self.assertEqual(param(cv), param(cvCopied))

        cvModel = cv.fit(dataset)
        cvModelCopied = cvModel.copy()
        for index in range(len(cvModel.avgMetrics)):
            self.assertTrue(
                abs(cvModel.avgMetrics[index] - cvModelCopied.avgMetrics[index]) < 0.0001
            )
        self.assertTrue(np.allclose(cvModel.stdMetrics, cvModelCopied.stdMetrics))
        # SPARK-32092: CrossValidatorModel.copy() needs to copy all existing params
        for param in [lambda x: x.getNumFolds(), lambda x: x.getFoldCol(), lambda x: x.getSeed()]:
            self.assertEqual(param(cvModel), param(cvModelCopied))

        cvModel.avgMetrics[0] = "foo"
        self.assertNotEqual(
            cvModelCopied.avgMetrics[0],
            "foo",
            "Changing the original avgMetrics should not affect the copied model",
        )
        cvModel.stdMetrics[0] = "foo"
        self.assertNotEqual(
            cvModelCopied.stdMetrics[0],
            "foo",
            "Changing the original stdMetrics should not affect the copied model",
        )

    def test_fit_minimize_metric(self):
        dataset = self.spark.createDataFrame(
            [(10, 10.0), (50, 50.0), (100, 100.0), (500, 500.0)] * 10, ["feature", "label"]
        )

        iee = InducedErrorEstimator()
        evaluator = RegressionEvaluator(metricName="rmse")

        grid = ParamGridBuilder().addGrid(iee.inducedError, [100.0, 0.0, 10000.0]).build()
        cv = CrossValidator(estimator=iee, estimatorParamMaps=grid, evaluator=evaluator)
        cvModel = cv.fit(dataset)
        bestModel = cvModel.bestModel
        bestModelMetric = evaluator.evaluate(bestModel.transform(dataset))

        self.assertEqual(
            0.0, bestModel.getOrDefault("inducedError"), "Best model should have zero induced error"
        )
        self.assertEqual(0.0, bestModelMetric, "Best model has RMSE of 0")

    def test_fit_maximize_metric(self):
        dataset = self.spark.createDataFrame(
            [(10, 10.0), (50, 50.0), (100, 100.0), (500, 500.0)] * 10, ["feature", "label"]
        )

        iee = InducedErrorEstimator()
        evaluator = RegressionEvaluator(metricName="r2")

        grid = ParamGridBuilder().addGrid(iee.inducedError, [100.0, 0.0, 10000.0]).build()
        cv = CrossValidator(estimator=iee, estimatorParamMaps=grid, evaluator=evaluator)
        cvModel = cv.fit(dataset)
        bestModel = cvModel.bestModel
        bestModelMetric = evaluator.evaluate(bestModel.transform(dataset))

        self.assertEqual(
            0.0, bestModel.getOrDefault("inducedError"), "Best model should have zero induced error"
        )
        self.assertEqual(1.0, bestModelMetric, "Best model has R-squared of 1")

    @staticmethod
    def _check_result(result_dataframe, expected_predictions, expected_probabilities=None):
        np.testing.assert_array_equal(list(result_dataframe.prediction), expected_predictions)
        if "probability" in result_dataframe.columns:
            np.testing.assert_allclose(
                list(result_dataframe.probability),
                expected_probabilities,
                rtol=1e-1,
            )

    def test_crossvalidator_on_pipeline(self):
        import torch
        from sklearn.datasets import load_breast_cancer

        sk_dataset = load_breast_cancer()

        train_dataset = self.spark.createDataFrame(
            zip(sk_dataset.data.tolist(), [int(t) for t in sk_dataset.target]),
            schema="features: array<double>, label: long",
        )

        scaler = StandardScaler(inputCol="features", outputCol="scaled_features")
        lorv2 = LORV2(numTrainWorkers=1, featuresCol="scaled_features")
        pipeline = Pipeline(stages=[scaler, lorv2])

        grid2 = ParamGridBuilder().addGrid(lorv2.maxIter, [2, 3]).build()
        cv = CrossValidator(
            estimator=pipeline,
            estimatorParamMaps=grid2,
            parallelism=2,
            evaluator=BinaryClassificationEvaluator(),
            numFolds=2,
        )
        cv_model = cv.fit(train_dataset)
        transformed_result = (
            cv_model.transform(train_dataset).select("prediction", "probability").toPandas()
        )
        expected_transformed_result = (
            cv_model.bestModel.transform(train_dataset)
            .select("prediction", "probability")
            .toPandas()
        )
        pd.testing.assert_frame_equal(transformed_result, expected_transformed_result)

        assert cv_model.bestModel.stages[1].getMaxIter() == 3

        # trial of index 2 should have better metric value
        # because it sets higher `maxIter` param.
        assert cv_model.avgMetrics[1] > cv_model.avgMetrics[0]

        def _verify_cv_saved_params(instance, loaded_instance):
            assert instance.getEstimator().uid == loaded_instance.getEstimator().uid
            assert instance.getEvaluator().uid == loaded_instance.getEvaluator().uid
            assert instance.getEstimatorParamMaps() == loaded_instance.getEstimatorParamMaps()

        # Test save / load
        with tempfile.TemporaryDirectory(prefix="test_crossvalidator_on_pipeline") as tmp_dir:
            with torch.serialization.safe_globals(
                [
                    torch.nn.modules.container.Sequential,
                    torch.nn.modules.linear.Linear,
                    torch.nn.modules.activation.Softmax,
                ]
            ):
                cv.saveToLocal(f"{tmp_dir}/cv")
                loaded_cv = CrossValidator.loadFromLocal(f"{tmp_dir}/cv")

                _verify_cv_saved_params(cv, loaded_cv)

                cv_model.saveToLocal(f"{tmp_dir}/cv_model")
                loaded_cv_model = CrossValidatorModel.loadFromLocal(f"{tmp_dir}/cv_model")

                _verify_cv_saved_params(cv_model, loaded_cv_model)

                assert cv_model.uid == loaded_cv_model.uid
                assert cv_model.bestModel.uid == loaded_cv_model.bestModel.uid
                assert cv_model.bestModel.stages[0].uid == loaded_cv_model.bestModel.stages[0].uid
                assert cv_model.bestModel.stages[1].uid == loaded_cv_model.bestModel.stages[1].uid
                assert loaded_cv_model.bestModel.stages[1].getMaxIter() == 3

                np.testing.assert_allclose(cv_model.avgMetrics, loaded_cv_model.avgMetrics)
                np.testing.assert_allclose(cv_model.stdMetrics, loaded_cv_model.stdMetrics)

    @unittest.skipIf(
        sys.version_info > (3, 12), "SPARK-46078: Fails with dev torch with Python 3.12"
    )
    def test_crossvalidator_with_fold_col(self):
        from sklearn.datasets import load_breast_cancer

        sk_dataset = load_breast_cancer()

        train_dataset = self.spark.createDataFrame(
            zip(
                sk_dataset.data.tolist(),
                [int(t) for t in sk_dataset.target],
                [int(i % 3) for i in range(len(sk_dataset.target))],
            ),
            schema="features: array<double>, label: long, fold: long",
        )

        lorv2 = LORV2(numTrainWorkers=2)

        grid2 = ParamGridBuilder().addGrid(lorv2.maxIter, [2, 3]).build()
        cv = CrossValidator(
            estimator=lorv2,
            estimatorParamMaps=grid2,
            parallelism=2,
            evaluator=BinaryClassificationEvaluator(),
            foldCol="fold",
            numFolds=3,
        )
        cv.fit(train_dataset)


@unittest.skipIf(
    not should_test_connect
    or not have_sklearn
    or not have_torch
    or not have_torcheval
    or is_remote_only(),
    connect_requirement_message
    or sklearn_requirement_message
    or torch_requirement_message
    or torcheval_requirement_message
    or "pyspark-connect cannot test classic Spark",
)
class CrossValidatorTests(CrossValidatorTestsMixin, ReusedSQLTestCase):
    @classmethod
    def master(cls):
        return "local[2]"


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import base64
from io import BytesIO
import unittest

import numpy as np
import pandas as pd

from pyspark import pandas as ps
from pyspark.pandas.config import set_option, reset_option
from pyspark.testing.pandasutils import PandasOnSparkTestCase, TestUtils
from pyspark.testing.utils import have_matplotlib, matplotlib_requirement_message

if have_matplotlib:
    import matplotlib
    from matplotlib import pyplot as plt

    matplotlib.use("agg")


@unittest.skipIf(not have_matplotlib, matplotlib_requirement_message)
class SeriesPlotMatplotlibTestsMixin:
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        pd.set_option("plotting.backend", "matplotlib")
        set_option("plotting.backend", "matplotlib")
        set_option("plotting.max_rows", 1000)

    @classmethod
    def tearDownClass(cls):
        pd.reset_option("plotting.backend")
        reset_option("plotting.backend")
        reset_option("plotting.max_rows")
        super().tearDownClass()

    @property
    def pdf1(self):
        return pd.DataFrame(
            {"a": [1, 2, 3, 4, 5, 6, 7, 8, 9, 15, 50]}, index=[0, 1, 3, 5, 6, 8, 9, 9, 9, 10, 10]
        )

    @property
    def psdf1(self):
        return ps.from_pandas(self.pdf1)

    @property
    def psdf2(self):
        return ps.range(1002)

    @property
    def pdf2(self):
        return self.psdf2._to_pandas()

    @staticmethod
    def plot_to_base64(ax):
        bytes_data = BytesIO()
        ax.figure.savefig(bytes_data, format="png")
        bytes_data.seek(0)
        b64_data = base64.b64encode(bytes_data.read())
        plt.close(ax.figure)
        return b64_data

    def test_bar_plot(self):
        pdf = self.pdf1
        psdf = self.psdf1

        ax1 = pdf["a"].plot(kind="bar", colormap="Paired")
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf["a"].plot(kind="bar", colormap="Paired")
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

        ax1 = pdf["a"].plot(kind="bar", colormap="Paired")
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf["a"].plot(kind="bar", colormap="Paired")
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

    def test_bar_plot_limited(self):
        pdf = self.pdf2
        psdf = self.psdf2

        _, ax1 = plt.subplots(1, 1)
        ax1 = pdf["id"][:1000].plot.bar(colormap="Paired")
        ax1.text(
            1,
            1,
            "showing top 1000 elements only",
            size=6,
            ha="right",
            va="bottom",
            transform=ax1.transAxes,
        )
        bin1 = self.plot_to_base64(ax1)

        _, ax2 = plt.subplots(1, 1)
        ax2 = psdf["id"].plot.bar(colormap="Paired")
        bin2 = self.plot_to_base64(ax2)

        self.assertEqual(bin1, bin2)

    def test_pie_plot(self):
        pdf = self.pdf1
        psdf = self.psdf1

        ax1 = pdf["a"].plot.pie(colormap="Paired")
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf["a"].plot.pie(colormap="Paired")
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

        ax1 = pdf["a"].plot(kind="pie", colormap="Paired")
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf["a"].plot(kind="pie", colormap="Paired")
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

    def test_pie_plot_limited(self):
        pdf = self.pdf2
        psdf = self.psdf2

        _, ax1 = plt.subplots(1, 1)
        ax1 = pdf["id"][:1000].plot.pie(colormap="Paired")
        ax1.text(
            1,
            1,
            "showing top 1000 elements only",
            size=6,
            ha="right",
            va="bottom",
            transform=ax1.transAxes,
        )
        bin1 = self.plot_to_base64(ax1)

        _, ax2 = plt.subplots(1, 1)
        ax2 = psdf["id"].plot.pie(colormap="Paired")
        bin2 = self.plot_to_base64(ax2)

        self.assertEqual(bin1, bin2)

    def test_line_plot(self):
        pdf = self.pdf1
        psdf = self.psdf1

        ax1 = pdf["a"].plot(kind="line", colormap="Paired")
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf["a"].plot(kind="line", colormap="Paired")
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

        ax1 = pdf["a"].plot.line(colormap="Paired")
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf["a"].plot.line(colormap="Paired")
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

    def test_barh_plot(self):
        pdf = self.pdf1
        psdf = self.psdf1

        ax1 = pdf["a"].plot(kind="barh", colormap="Paired")
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf["a"].plot(kind="barh", colormap="Paired")
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

    def test_barh_plot_limited(self):
        pdf = self.pdf2
        psdf = self.psdf2

        _, ax1 = plt.subplots(1, 1)
        ax1 = pdf["id"][:1000].plot.barh(colormap="Paired")
        ax1.text(
            1,
            1,
            "showing top 1000 elements only",
            size=6,
            ha="right",
            va="bottom",
            transform=ax1.transAxes,
        )
        bin1 = self.plot_to_base64(ax1)

        _, ax2 = plt.subplots(1, 1)
        ax2 = psdf["id"].plot.barh(colormap="Paired")
        bin2 = self.plot_to_base64(ax2)

        self.assertEqual(bin1, bin2)

    def test_hist(self):
        pdf = pd.DataFrame(
            {"a": [1, 2, 3, 4, 5, 6, 7, 8, 9, 15, 50]}, index=[0, 1, 3, 5, 6, 8, 9, 9, 9, 10, 10]
        )

        psdf = ps.from_pandas(pdf)

        def plot_to_base64(ax):
            bytes_data = BytesIO()
            ax.figure.savefig(bytes_data, format="png")
            bytes_data.seek(0)
            b64_data = base64.b64encode(bytes_data.read())
            plt.close(ax.figure)
            return b64_data

        _, ax1 = plt.subplots(1, 1)
        # Using plot.hist() because pandas changes ticks props when called hist()
        ax1 = pdf["a"].plot.hist()
        _, ax2 = plt.subplots(1, 1)
        ax2 = psdf["a"].hist()
        self.assert_eq(plot_to_base64(ax1), plot_to_base64(ax2))

    def test_hist_plot(self):
        pdf = self.pdf1
        psdf = self.psdf1

        _, ax1 = plt.subplots(1, 1)
        ax1 = pdf["a"].plot.hist()
        bin1 = self.plot_to_base64(ax1)
        _, ax2 = plt.subplots(1, 1)
        ax2 = psdf["a"].plot.hist()
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

        ax1 = pdf["a"].plot.hist(bins=15)
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf["a"].plot.hist(bins=15)
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

        ax1 = pdf["a"].plot(kind="hist", bins=15)
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf["a"].plot(kind="hist", bins=15)
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

        ax1 = pdf["a"].plot.hist(bins=3, bottom=[2, 1, 3])
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf["a"].plot.hist(bins=3, bottom=[2, 1, 3])
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

    def test_area_plot(self):
        pdf = pd.DataFrame(
            {
                "sales": [3, 2, 3, 9, 10, 6],
                "signups": [5, 5, 6, 12, 14, 13],
                "visits": [20, 42, 28, 62, 81, 50],
            },
            index=pd.date_range(start="2018/01/01", end="2018/07/01", freq="ME"),
        )
        psdf = ps.from_pandas(pdf)

        ax1 = pdf["sales"].plot(kind="area", colormap="Paired")
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf["sales"].plot(kind="area", colormap="Paired")
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

        ax1 = pdf["sales"].plot.area(colormap="Paired")
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf["sales"].plot.area(colormap="Paired")
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

        # just a sanity check for df.col type
        ax1 = pdf.sales.plot(kind="area", colormap="Paired")
        bin1 = self.plot_to_base64(ax1)
        ax2 = psdf.sales.plot(kind="area", colormap="Paired")
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)

    def test_box_plot(self):
        def check_box_plot(pser, psser, *args, **kwargs):
            _, ax1 = plt.subplots(1, 1)
            ax1 = pser.plot.box(*args, **kwargs)
            _, ax2 = plt.subplots(1, 1)
            ax2 = psser.plot.box(*args, **kwargs)

            diffs = [
                np.array([0, 0.5, 0, 0.5, 0, -0.5, 0, -0.5, 0, 0.5]),
                np.array([0, 0.5, 0, 0]),
                np.array([0, -0.5, 0, 0]),
            ]

            try:
                for i, (line1, line2) in enumerate(zip(ax1.get_lines(), ax2.get_lines())):
                    expected = line1.get_xydata().ravel()
                    actual = line2.get_xydata().ravel()
                    if i < 3:
                        actual += diffs[i]
                    self.assert_eq(pd.Series(expected), pd.Series(actual))
            finally:
                ax1.cla()
                ax2.cla()

        # Non-named Series
        pser = pd.Series([1, 2, 3, 4, 5, 6, 7, 8, 9, 15, 50], [0, 1, 3, 5, 6, 8, 9, 9, 9, 10, 10])
        psser = ps.from_pandas(pser)

        spec = [(self.pdf1.a, self.psdf1.a), (pser, psser)]

        for p, k in spec:
            check_box_plot(p, k)
            check_box_plot(p, k, showfliers=True)
            check_box_plot(p, k, sym="")
            check_box_plot(p, k, sym=".", color="r")
            check_box_plot(p, k, use_index=False)
            check_box_plot(p, k, usermedians=[2.0])
            check_box_plot(p, k, conf_intervals=[(1.0, 3.0)])

        val = (1, 3)
        self.assertRaises(
            ValueError, lambda: check_box_plot(self.pdf1, self.psdf1, usermedians=[2.0, 3.0])
        )
        self.assertRaises(
            ValueError, lambda: check_box_plot(self.pdf1, self.psdf1, conf_intervals=[val, val])
        )
        self.assertRaises(
            ValueError, lambda: check_box_plot(self.pdf1, self.psdf1, conf_intervals=[(1,)])
        )

    def test_kde_plot(self):
        def moving_average(a, n=10):
            ret = np.cumsum(a, dtype=float)
            ret[n:] = ret[n:] - ret[:-n]
            return ret[n - 1 :] / n

        def check_kde_plot(pdf, psdf, *args, **kwargs):
            _, ax1 = plt.subplots(1, 1)
            ax1 = pdf["a"].plot.kde(*args, **kwargs)
            _, ax2 = plt.subplots(1, 1)
            ax2 = psdf["a"].plot.kde(*args, **kwargs)

            try:
                for i, (line1, line2) in enumerate(zip(ax1.get_lines(), ax2.get_lines())):
                    expected = line1.get_xydata().ravel()
                    actual = line2.get_xydata().ravel()
                    # TODO: Due to implementation difference, the output is different comparing
                    # to pandas'. We should identify the root cause of difference, and reduce
                    # the diff.

                    # Note: Data is from 1 to 50. So, it smooths them by moving average and compares
                    # both.
                    self.assertTrue(
                        np.allclose(moving_average(actual), moving_average(expected), rtol=3)
                    )
            finally:
                ax1.cla()
                ax2.cla()

        check_kde_plot(self.pdf1, self.psdf1, bw_method=0.3)
        check_kde_plot(self.pdf1, self.psdf1, ind=[1, 2, 3, 4, 5], bw_method=3.0)

    def test_empty_hist(self):
        pdf = self.pdf1.assign(categorical="A")
        psdf = ps.from_pandas(pdf)
        psser = psdf["categorical"]

        with self.assertRaisesRegex(TypeError, "Empty 'DataFrame': no numeric data to plot"):
            psser.plot.hist()

    def test_single_value_hist(self):
        pdf = self.pdf1.assign(single=2)
        psdf = ps.from_pandas(pdf)

        _, ax1 = plt.subplots(1, 1)
        ax1 = pdf["single"].plot.hist()
        bin1 = self.plot_to_base64(ax1)
        _, ax2 = plt.subplots(1, 1)
        ax2 = psdf["single"].plot.hist()
        bin2 = self.plot_to_base64(ax2)
        self.assertEqual(bin1, bin2)


class SeriesPlotMatplotlibTests(SeriesPlotMatplotlibTestsMixin, PandasOnSparkTestCase, TestUtils):
    pass


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tempfile
import unittest

from pyspark.sql import Row
from pyspark.ml.pipeline import Pipeline, PipelineModel
from pyspark.ml.feature import (
    VectorAssembler,
    MaxAbsScaler,
    MaxAbsScalerModel,
    MinMaxScaler,
    MinMaxScalerModel,
)
from pyspark.ml.linalg import Vectors
from pyspark.ml.classification import LogisticRegression, LogisticRegressionModel
from pyspark.ml.clustering import KMeans, KMeansModel, GaussianMixture
from pyspark.testing.mlutils import MockDataset, MockEstimator, MockTransformer
from pyspark.testing.sqlutils import ReusedSQLTestCase


class PipelineTestsMixin:
    def test_pipeline(self):
        dataset = MockDataset()
        estimator0 = MockEstimator()
        transformer1 = MockTransformer()
        estimator2 = MockEstimator()
        transformer3 = MockTransformer()
        pipeline = Pipeline(stages=[estimator0, transformer1, estimator2, transformer3])
        pipeline_model = pipeline.fit(dataset, {estimator0.fake: 0, transformer1.fake: 1})
        model0, transformer1, model2, transformer3 = pipeline_model.stages
        self.assertEqual(0, model0.dataset_index)
        self.assertEqual(0, model0.getFake())
        self.assertEqual(1, transformer1.dataset_index)
        self.assertEqual(1, transformer1.getFake())
        self.assertEqual(2, dataset.index)
        self.assertIsNone(model2.dataset_index, "The last model shouldn't be called in fit.")
        self.assertIsNone(
            transformer3.dataset_index, "The last transformer shouldn't be called in fit."
        )
        dataset = pipeline_model.transform(dataset)
        self.assertEqual(2, model0.dataset_index)
        self.assertEqual(3, transformer1.dataset_index)
        self.assertEqual(4, model2.dataset_index)
        self.assertEqual(5, transformer3.dataset_index)
        self.assertEqual(6, dataset.index)

    def test_identity_pipeline(self):
        dataset = MockDataset()

        def doTransform(pipeline):
            pipeline_model = pipeline.fit(dataset)
            return pipeline_model.transform(dataset)

        # check that empty pipeline did not perform any transformation
        self.assertEqual(dataset.index, doTransform(Pipeline(stages=[])).index)
        # check that failure to set stages param will raise KeyError for missing param
        self.assertRaises(KeyError, lambda: doTransform(Pipeline()))

    def test_classification_pipeline(self):
        df = self.spark.createDataFrame([(1, 1, 0, 3), (0, 2, 0, 1)], ["y", "a", "b", "c"])

        assembler = VectorAssembler(outputCol="features")
        assembler.setInputCols(["a", "b", "c"])

        scaler = MaxAbsScaler(inputCol="features", outputCol="scaled_features")

        lr = LogisticRegression(featuresCol="scaled_features", labelCol="y")
        lr.setMaxIter(1)

        pipeline = Pipeline(stages=[assembler, scaler, lr])
        self.assertEqual(len(pipeline.getStages()), 3)

        model = pipeline.fit(df)
        self.assertEqual(len(model.stages), 3)
        self.assertIsInstance(model.stages[0], VectorAssembler)
        self.assertIsInstance(model.stages[1], MaxAbsScalerModel)
        self.assertIsInstance(model.stages[2], LogisticRegressionModel)

        output = model.transform(df)
        self.assertEqual(
            output.columns,
            [
                "y",
                "a",
                "b",
                "c",
                "features",
                "scaled_features",
                "rawPrediction",
                "probability",
                "prediction",
            ],
        )
        self.assertEqual(output.count(), 2)

        # save & load
        with tempfile.TemporaryDirectory(prefix="classification_pipeline") as d:
            pipeline.write().overwrite().save(d)
            pipeline2 = Pipeline.load(d)
            self.assertEqual(str(pipeline), str(pipeline2))
            self.assertEqual(str(pipeline.getStages()), str(pipeline2.getStages()))

            model.write().overwrite().save(d)
            model2 = PipelineModel.load(d)
            self.assertEqual(str(model), str(model2))
            self.assertEqual(str(model.stages), str(model2.stages))

    def test_clustering_pipeline(self):
        df = self.spark.createDataFrame([(1, 1, 0, 3), (0, 2, 0, 1)], ["y", "a", "b", "c"])

        assembler = VectorAssembler(outputCol="features")
        assembler.setInputCols(["a", "b", "c"])

        scaler = MinMaxScaler(inputCol="features", outputCol="scaled_features")

        km = KMeans(k=2, maxIter=2)

        pipeline = Pipeline(stages=[assembler, scaler, km])
        self.assertEqual(len(pipeline.getStages()), 3)

        # Pipeline save & load
        with tempfile.TemporaryDirectory(prefix="clustering_pipeline") as d:
            pipeline.write().overwrite().save(d)
            pipeline2 = Pipeline.load(d)
            self.assertEqual(str(pipeline), str(pipeline2))
            self.assertEqual(str(pipeline.getStages()), str(pipeline2.getStages()))

        model = pipeline.fit(df)
        self.assertEqual(len(model.stages), 3)
        self.assertIsInstance(model.stages[0], VectorAssembler)
        self.assertIsInstance(model.stages[1], MinMaxScalerModel)
        self.assertIsInstance(model.stages[2], KMeansModel)

        output = model.transform(df)
        self.assertEqual(
            output.columns,
            [
                "y",
                "a",
                "b",
                "c",
                "features",
                "scaled_features",
                "prediction",
            ],
        )
        self.assertEqual(output.count(), 2)

        # PipelineModel save & load
        with tempfile.TemporaryDirectory(prefix="clustering_pipeline") as d:
            pipeline.write().overwrite().save(d)
            pipeline2 = Pipeline.load(d)
            self.assertEqual(str(pipeline), str(pipeline2))
            self.assertEqual(str(pipeline.getStages()), str(pipeline2.getStages()))

            model.write().overwrite().save(d)
            model2 = PipelineModel.load(d)
            self.assertEqual(str(model), str(model2))
            self.assertEqual(str(model.stages), str(model2.stages))

    @unittest.skip("Disabled due to flakiness, it might hang forever occasionally.")
    def test_model_gc(self):
        spark = self.spark
        df1 = spark.createDataFrame(
            [
                Row(label=0.0, weight=0.1, features=Vectors.dense([0.0, 0.0])),
                Row(label=0.0, weight=0.5, features=Vectors.dense([0.0, 1.0])),
                Row(label=1.0, weight=1.0, features=Vectors.dense([1.0, 0.0])),
            ]
        )

        def fit_transform(df):
            lr = LogisticRegression(maxIter=1, regParam=0.01, weightCol="weight")
            model = lr.fit(df)
            return model.transform(df)

        output1 = fit_transform(df1)
        self.assertEqual(output1.count(), 3)

        df2 = spark.range(10)

        def fit_transform_and_union(df1, df2):
            output1 = fit_transform(df1)
            return output1.unionByName(df2, True)

        output2 = fit_transform_and_union(df1, df2)
        self.assertEqual(output2.count(), 13)

    @unittest.skip("Disabled due to flakiness, it might hang forever occasionally.")
    def test_model_training_summary_gc(self):
        spark = self.spark
        df1 = spark.createDataFrame(
            [
                Row(label=0.0, weight=0.1, features=Vectors.dense([0.0, 0.0])),
                Row(label=0.0, weight=0.5, features=Vectors.dense([0.0, 1.0])),
                Row(label=1.0, weight=1.0, features=Vectors.dense([1.0, 0.0])),
            ]
        )

        def fit_predictions(df):
            lr = LogisticRegression(maxIter=1, regParam=0.01, weightCol="weight")
            model = lr.fit(df)
            return model.summary.predictions

        output1 = fit_predictions(df1)
        self.assertEqual(output1.count(), 3)

        df2 = spark.range(10)

        def fit_predictions_and_union(df1, df2):
            output1 = fit_predictions(df1)
            return output1.unionByName(df2, True)

        output2 = fit_predictions_and_union(df1, df2)
        self.assertEqual(output2.count(), 13)

    @unittest.skip("Disabled due to flakiness, it might hang forever occasionally.")
    def test_model_testing_summary_gc(self):
        spark = self.spark
        df1 = spark.createDataFrame(
            [
                Row(label=0.0, weight=0.1, features=Vectors.dense([0.0, 0.0])),
                Row(label=0.0, weight=0.5, features=Vectors.dense([0.0, 1.0])),
                Row(label=1.0, weight=1.0, features=Vectors.dense([1.0, 0.0])),
            ]
        )

        def fit_predictions(df):
            lr = LogisticRegression(maxIter=1, regParam=0.01, weightCol="weight")
            model = lr.fit(df)
            return model.evaluate(df).predictions

        output1 = fit_predictions(df1)
        self.assertEqual(output1.count(), 3)

        df2 = spark.range(10)

        def fit_predictions_and_union(df1, df2):
            output1 = fit_predictions(df1)
            return output1.unionByName(df2, True)

        output2 = fit_predictions_and_union(df1, df2)
        self.assertEqual(output2.count(), 13)

    @unittest.skip("Disabled due to flakiness, it might hang forever occasionally.")
    def test_model_attr_df_gc(self):
        spark = self.spark
        df1 = (
            spark.createDataFrame(
                [
                    (1, 1.0, Vectors.dense([-0.1, -0.05])),
                    (2, 2.0, Vectors.dense([-0.01, -0.1])),
                    (3, 3.0, Vectors.dense([0.9, 0.8])),
                    (4, 1.0, Vectors.dense([0.75, 0.935])),
                    (5, 1.0, Vectors.dense([-0.83, -0.68])),
                    (6, 1.0, Vectors.dense([-0.91, -0.76])),
                ],
                ["index", "weight", "features"],
            )
            .coalesce(1)
            .sortWithinPartitions("index")
            .select("weight", "features")
        )

        def fit_attr_df(df):
            gmm = GaussianMixture(k=2, maxIter=2, weightCol="weight", seed=1)
            model = gmm.fit(df)
            return model.gaussiansDF

        output1 = fit_attr_df(df1)
        self.assertEqual(output1.count(), 2)

        df2 = spark.range(10)

        def fit_attr_df_and_union(df1, df2):
            output1 = fit_attr_df(df1)
            return output1.unionByName(df2, True)

        output2 = fit_attr_df_and_union(df1, df2)
        self.assertEqual(output2.count(), 12)


class PipelineTests(PipelineTestsMixin, ReusedSQLTestCase):
    pass


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Utilities to deal with types. This is mostly focused on python3.
"""

import datetime
import decimal
import sys
import typing
from collections.abc import Iterable
from inspect import isclass
from typing import Any, Callable, Generic, List, Tuple, Union, Type, get_type_hints

import numpy as np
import pandas as pd
from pandas.api.types import CategoricalDtype, pandas_dtype
from pandas.api.extensions import ExtensionDtype

from pyspark.loose_version import LooseVersion

extension_dtypes: Tuple[type, ...]
try:
    from pandas import Int8Dtype, Int16Dtype, Int32Dtype, Int64Dtype

    extension_dtypes_available = True
    extension_dtypes = (Int8Dtype, Int16Dtype, Int32Dtype, Int64Dtype)

    try:
        from pandas import BooleanDtype, StringDtype

        extension_object_dtypes_available = True
        extension_dtypes += (BooleanDtype, StringDtype)
    except ImportError:
        extension_object_dtypes_available = False

    try:
        from pandas import Float32Dtype, Float64Dtype

        extension_float_dtypes_available = True
        extension_dtypes += (Float32Dtype, Float64Dtype)
    except ImportError:
        extension_float_dtypes_available = False

except ImportError:
    extension_dtypes_available = False
    extension_object_dtypes_available = False
    extension_float_dtypes_available = False
    extension_dtypes = ()

import pyarrow as pa
import pyspark.sql.types as types
from pyspark.sql.pandas.types import to_arrow_type, from_arrow_type

# For running doctests and reference resolution in PyCharm.
from pyspark import pandas as ps  # noqa: F401
from pyspark.pandas._typing import Dtype, T

if typing.TYPE_CHECKING:
    from pyspark.pandas.internal import InternalField


# A column of data, with the data type.
class SeriesType(Generic[T]):
    def __init__(self, dtype: Dtype, spark_type: types.DataType):
        self.dtype = dtype
        self.spark_type = spark_type

    def __repr__(self) -> str:
        return "SeriesType[{}]".format(self.spark_type)


class DataFrameType:
    def __init__(
        self,
        index_fields: List["InternalField"],
        data_fields: List["InternalField"],
    ):
        self.index_fields = index_fields
        self.data_fields = data_fields
        self.fields = index_fields + data_fields

    @property
    def dtypes(self) -> List[Dtype]:
        return [field.dtype for field in self.fields]

    @property
    def spark_type(self) -> types.StructType:
        return types.StructType([field.struct_field for field in self.fields])

    def __repr__(self) -> str:
        return "DataFrameType[{}]".format(self.spark_type)


# The type is a scalar type that is furthermore understood by Spark.
class ScalarType:
    def __init__(self, dtype: Dtype, spark_type: types.DataType):
        self.dtype = dtype
        self.spark_type = spark_type

    def __repr__(self) -> str:
        return "ScalarType[{}]".format(self.spark_type)


# The type is left unspecified or we do not know about this type.
class UnknownType:
    def __init__(self, tpe: Any):
        self.tpe = tpe

    def __repr__(self) -> str:
        return "UnknownType[{}]".format(self.tpe)


class IndexNameTypeHolder:
    name = None
    tpe = None
    short_name = "IndexNameType"


class NameTypeHolder:
    name = None
    tpe = None
    short_name = "NameType"


def as_spark_type(
    tpe: Union[str, type, Dtype], *, raise_error: bool = True, prefer_timestamp_ntz: bool = False
) -> types.DataType:
    """
    Given a Python type, returns the equivalent spark type.
    Accepts:
    - the built-in types in Python
    - the built-in types in numpy
    - list of pairs of (field_name, type)
    - dictionaries of field_name -> type
    - Python3's typing system
    """
    # For NumPy typing, NumPy version should be 1.21+
    if LooseVersion(np.__version__) >= LooseVersion("1.21"):
        if (
            hasattr(tpe, "__origin__")
            and tpe.__origin__ is np.ndarray
            and hasattr(tpe, "__args__")
            and len(tpe.__args__) > 1
        ):
            # numpy.typing.NDArray for numpy < 2.5
            return types.ArrayType(
                as_spark_type(tpe.__args__[1].__args__[0], raise_error=raise_error)
            )
        elif (
            hasattr(tpe, "__origin__")
            and hasattr(tpe.__origin__, "__value__")
            and getattr(tpe.__origin__.__value__, "__origin__", None) is np.ndarray
            and hasattr(tpe, "__args__")
            and len(tpe.__args__) > 0
        ):
            # numpy.typing.NDArray for numpy >= 2.5: a PEP 695 type alias whose __value__
            # resolves to np.ndarray[shape, dtype[scalar]], with the scalar at __args__[0]
            return types.ArrayType(as_spark_type(tpe.__args__[0], raise_error=raise_error))

    if isinstance(tpe, np.dtype) and tpe == np.dtype("object"):
        pass
    # ArrayType
    elif tpe in (np.ndarray,):
        return types.ArrayType(types.StringType())
    elif hasattr(tpe, "__origin__") and issubclass(tpe.__origin__, list):
        element_type = as_spark_type(
            tpe.__args__[0],  # type: ignore[union-attr]
            raise_error=raise_error,
        )
        if element_type is None:
            return None
        return types.ArrayType(element_type)
    # BinaryType
    elif tpe in (bytes, np.character, np.bytes_):
        return types.BinaryType()
    # BooleanType
    elif tpe in (bool, np.bool_, "bool", "?"):
        return types.BooleanType()
    # DateType
    elif tpe in (datetime.date,):
        return types.DateType()
    # NumericType
    elif tpe in (np.int8, np.byte, "int8", "byte", "b"):
        return types.ByteType()
    elif tpe in (decimal.Decimal,):
        # TODO: considering the precision & scale for decimal type.
        return types.DecimalType(38, 18)
    elif tpe in (float, np.double, np.float64, "float", "float64", "double"):
        return types.DoubleType()
    elif tpe in (np.float32, "float32", "f"):
        return types.FloatType()
    elif tpe in (np.int32, "int32", "i"):
        return types.IntegerType()
    elif tpe in (int, np.int64, "int", "int64", "long"):
        return types.LongType()
    elif tpe in (np.int16, "int16", "short"):
        return types.ShortType()
    # StringType
    elif tpe in (str, np.str_, "str", "U"):
        return types.StringType()
    # TimestampType or TimestampNTZType if timezone is not specified.
    elif tpe in (datetime.datetime, np.datetime64, "M", pd.Timestamp) or (
        isinstance(tpe, np.dtype) and tpe.type is np.datetime64
    ):
        return types.TimestampNTZType() if prefer_timestamp_ntz else types.TimestampType()

    # DayTimeIntervalType
    elif tpe in (datetime.timedelta, np.timedelta64) or (
        isinstance(tpe, np.dtype) and tpe.type is np.timedelta64
    ):
        return types.DayTimeIntervalType()

    # categorical types
    elif isinstance(tpe, CategoricalDtype) or (isinstance(tpe, str) and tpe == "category"):
        return types.LongType()

    # extension types
    elif extension_dtypes_available:
        # IntegralType
        if isinstance(tpe, Int8Dtype) or (isinstance(tpe, str) and tpe == "Int8"):
            return types.ByteType()
        elif isinstance(tpe, Int16Dtype) or (isinstance(tpe, str) and tpe == "Int16"):
            return types.ShortType()
        elif isinstance(tpe, Int32Dtype) or (isinstance(tpe, str) and tpe == "Int32"):
            return types.IntegerType()
        elif isinstance(tpe, Int64Dtype) or (isinstance(tpe, str) and tpe == "Int64"):
            return types.LongType()

        if extension_object_dtypes_available:
            # BooleanType
            if isinstance(tpe, BooleanDtype) or (isinstance(tpe, str) and tpe == "boolean"):
                return types.BooleanType()
            # StringType
            elif isinstance(tpe, StringDtype) or (isinstance(tpe, str) and tpe == "string"):
                return types.StringType()

        if extension_float_dtypes_available:
            # FractionalType
            if isinstance(tpe, Float32Dtype) or (isinstance(tpe, str) and tpe == "Float32"):
                return types.FloatType()
            elif isinstance(tpe, Float64Dtype) or (isinstance(tpe, str) and tpe == "Float64"):
                return types.DoubleType()

    if raise_error:
        raise TypeError("Type %s was not understood." % tpe)
    else:
        return None


def spark_type_to_pandas_dtype(
    spark_type: types.DataType, *, use_extension_dtypes: bool = False
) -> Dtype:
    """Return the given Spark DataType to pandas dtype."""

    if use_extension_dtypes and extension_dtypes_available:
        # IntegralType
        if isinstance(spark_type, types.ByteType):
            return Int8Dtype()
        elif isinstance(spark_type, types.ShortType):
            return Int16Dtype()
        elif isinstance(spark_type, types.IntegerType):
            return Int32Dtype()
        elif isinstance(spark_type, types.LongType):
            return Int64Dtype()

        if extension_object_dtypes_available:
            # BooleanType
            if isinstance(spark_type, types.BooleanType):
                return BooleanDtype()
            # StringType
            elif isinstance(spark_type, types.StringType):
                return StringDtype()

        # FractionalType
        if extension_float_dtypes_available:
            if isinstance(spark_type, types.FloatType):
                return Float32Dtype()
            elif isinstance(spark_type, types.DoubleType):
                return Float64Dtype()

    if LooseVersion(pd.__version__) >= "3.0.0":
        if extension_object_dtypes_available and isinstance(spark_type, types.StringType):
            return StringDtype(na_value=np.nan)

    if isinstance(
        spark_type,
        (
            types.DateType,
            types.NullType,
            types.ArrayType,
            types.MapType,
            types.StructType,
            types.UserDefinedType,
        ),
    ):
        return np.dtype("object")
    elif isinstance(spark_type, types.DayTimeIntervalType):
        if LooseVersion(pd.__version__) < "3.0.0":
            return np.dtype("timedelta64[ns]")
        else:
            return np.dtype("timedelta64[us]")
    elif isinstance(spark_type, (types.TimestampType, types.TimestampNTZType)):
        if LooseVersion(pd.__version__) < "3.0.0":
            return np.dtype("datetime64[ns]")
        else:
            return np.dtype("datetime64[us]")
    else:
        from pyspark.pandas.utils import default_session

        prefers_large_var_types = (
            default_session()
            .conf.get("spark.sql.execution.arrow.useLargeVarTypes", "false")
            .lower()
            == "true"
        )
        return np.dtype(
            to_arrow_type(
                spark_type, timezone="UTC", prefers_large_types=prefers_large_var_types
            ).to_pandas_dtype()
        )


def is_str_dtype(tpe: Dtype) -> bool:
    if LooseVersion(pd.__version__) < "3.0.0":
        return False
    if extension_object_dtypes_available:
        return isinstance(tpe, StringDtype) and tpe.na_value is np.nan
    return False


def handle_dtype_as_extension_dtype(tpe: Dtype) -> bool:
    if is_str_dtype(tpe):
        return False
    else:
        return isinstance(tpe, extension_dtypes)


def pandas_on_spark_type(tpe: Union[str, type, Dtype]) -> Tuple[Dtype, types.DataType]:
    """
    Convert input into a pandas only dtype object or a numpy dtype object,
    and its corresponding Spark DataType.

    Parameters
    ----------
    tpe : object to be converted

    Returns
    -------
    tuple of np.dtype or a pandas dtype, and Spark DataType

    Raises
    ------
    TypeError if not a dtype

    Examples
    --------
    >>> from pyspark.loose_version import LooseVersion
    >>> using_pandas_3 = LooseVersion(pd.__version__) >= "3.0.0"

    >>> pandas_on_spark_type(int)
    (dtype('int64'), LongType())
    >>> t = pandas_on_spark_type(str)
    >>> if using_pandas_3:
    ...     t[0] == pd.StringDtype(na_value=np.nan)
    ... else:
    ...     t[0] == np.str_
    True
    >>> t[1]
    StringType()
    >>> pandas_on_spark_type(datetime.date)
    (dtype('O'), DateType())
    >>> pandas_on_spark_type(datetime.datetime)
    (dtype('<M8[ns]'), TimestampType())
    >>> pandas_on_spark_type(datetime.timedelta)
    (dtype('<m8[ns]'), DayTimeIntervalType(0, 3))
    >>> pandas_on_spark_type(List[bool])
    (dtype('O'), ArrayType(BooleanType(), True))
    """
    try:
        dtype = pandas_dtype(tpe)
        spark_type = as_spark_type(dtype)
    except (TypeError, ValueError):
        spark_type = as_spark_type(tpe)
        dtype = spark_type_to_pandas_dtype(spark_type)
    return dtype, spark_type


def infer_pd_series_spark_type(
    pser: pd.Series, dtype: Dtype, prefer_timestamp_ntz: bool = False
) -> types.DataType:
    """Infer Spark DataType from pandas Series dtype.

    :param pser: :class:`pandas.Series` to be inferred
    :param dtype: the Series' dtype
    :param prefer_timestamp_ntz: if true, infers datetime without timezone as
        TimestampNTZType type. If false, infers it as TimestampType.
    :return: the inferred Spark data type
    """
    if dtype == np.dtype("object"):
        if len(pser) == 0 or pser.isnull().all():
            return types.NullType()
        notnull = pser[pser.notnull()]
        if hasattr(notnull.iloc[0], "__UDT__"):
            return notnull.iloc[0].__UDT__
        else:
            return from_arrow_type(pa.Array.from_pandas(pser).type, prefer_timestamp_ntz)
    elif isinstance(dtype, CategoricalDtype):
        if isinstance(pser.dtype, CategoricalDtype):
            return as_spark_type(pser.cat.codes.dtype, prefer_timestamp_ntz=prefer_timestamp_ntz)
        else:
            # `pser` must already be converted to codes.
            return as_spark_type(pser.dtype, prefer_timestamp_ntz=prefer_timestamp_ntz)
    else:
        return as_spark_type(dtype, prefer_timestamp_ntz=prefer_timestamp_ntz)


def infer_return_type(f: Callable) -> Union[SeriesType, DataFrameType, ScalarType, UnknownType]:
    """
    Infer the return type from the return type annotation of the given function.

    The returned type class indicates both dtypes (a pandas only dtype object
    or a numpy dtype object) and its corresponding Spark DataType.

    >>> from pyspark.loose_version import LooseVersion
    >>> using_pandas_3 = LooseVersion(pd.__version__) >= "3.0.0"

    >>> def func() -> int:
    ...    pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtype
    dtype('int64')
    >>> inferred.spark_type
    LongType()

    >>> def func() -> ps.Series[int]:
    ...    pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtype
    dtype('int64')
    >>> inferred.spark_type
    LongType()

    >>> def func() -> ps.DataFrame[float, str]:
    ...    pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes[0]
    dtype('float64')
    >>> if using_pandas_3:
    ...     inferred.dtypes[1] == pd.StringDtype(na_value=np.nan)
    ... else:
    ...     inferred.dtypes[1] == np.str_
    True
    >>> inferred.spark_type
    StructType([StructField('c0', DoubleType(), True), StructField('c1', StringType(), True)])

    >>> def func() -> ps.DataFrame[float]:
    ...    pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [dtype('float64')]
    >>> inferred.spark_type
    StructType([StructField('c0', DoubleType(), True)])

    >>> def func() -> 'int':
    ...    pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtype
    dtype('int64')
    >>> inferred.spark_type
    LongType()

    >>> def func() -> 'ps.Series[int]':
    ...    pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtype
    dtype('int64')
    >>> inferred.spark_type
    LongType()

    >>> def func() -> 'ps.DataFrame[float, str]':
    ...    pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes[0]
    dtype('float64')
    >>> if using_pandas_3:
    ...     inferred.dtypes[1] == pd.StringDtype(na_value=np.nan)
    ... else:
    ...     inferred.dtypes[1] == np.str_
    True
    >>> inferred.spark_type
    StructType([StructField('c0', DoubleType(), True), StructField('c1', StringType(), True)])

    >>> def func() -> 'ps.DataFrame[float]':
    ...    pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [dtype('float64')]
    >>> inferred.spark_type
    StructType([StructField('c0', DoubleType(), True)])

    >>> def func() -> ps.DataFrame['a': float, 'b': int]:
    ...     pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [dtype('float64'), dtype('int64')]
    >>> inferred.spark_type
    StructType([StructField('a', DoubleType(), True), StructField('b', LongType(), True)])

    >>> def func() -> "ps.DataFrame['a': float, 'b': int]":
    ...     pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [dtype('float64'), dtype('int64')]
    >>> inferred.spark_type
    StructType([StructField('a', DoubleType(), True), StructField('b', LongType(), True)])

    >>> pdf = pd.DataFrame({"a": [1, 2, 3], "b": [3, 4, 5]})
    >>> def func() -> ps.DataFrame[pdf.dtypes]:
    ...     pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [dtype('int64'), dtype('int64')]
    >>> inferred.spark_type
    StructType([StructField('c0', LongType(), True), StructField('c1', LongType(), True)])

    >>> pdf = pd.DataFrame({"a": [1, 2, 3], "b": [3, 4, 5]})
    >>> def func() -> ps.DataFrame[zip(pdf.columns, pdf.dtypes)]:
    ...     pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [dtype('int64'), dtype('int64')]
    >>> inferred.spark_type
    StructType([StructField('a', LongType(), True), StructField('b', LongType(), True)])

    >>> pdf = pd.DataFrame({("x", "a"): [1, 2, 3], ("y", "b"): [3, 4, 5]})
    >>> def func() -> ps.DataFrame[zip(pdf.columns, pdf.dtypes)]:
    ...     pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [dtype('int64'), dtype('int64')]
    >>> inferred.spark_type
    StructType([StructField('(x, a)', LongType(), True), StructField('(y, b)', LongType(), True)])

    >>> pdf = pd.DataFrame({"a": [1, 2, 3], "b": pd.Categorical([3, 4, 5])})
    >>> def func() -> ps.DataFrame[pdf.dtypes]:
    ...     pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [dtype('int64'), CategoricalDtype(categories=[3, 4, 5], ordered=False, categories_dtype=int64)]
    >>> inferred.spark_type
    StructType([StructField('c0', LongType(), True), StructField('c1', LongType(), True)])

    >>> def func() -> ps.DataFrame[zip(pdf.columns, pdf.dtypes)]:
    ...     pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [dtype('int64'), CategoricalDtype(categories=[3, 4, 5], ordered=False, categories_dtype=int64)]
    >>> inferred.spark_type
    StructType([StructField('a', LongType(), True), StructField('b', LongType(), True)])

    >>> def func() -> ps.Series[pdf.b.dtype]:
    ...     pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtype
    CategoricalDtype(categories=[3, 4, 5], ordered=False, categories_dtype=int64)
    >>> inferred.spark_type
    LongType()

    >>> def func() -> ps.DataFrame[int, [int, int]]:
    ...     pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [dtype('int64'), dtype('int64'), dtype('int64')]
    >>> inferred.spark_type.simpleString()
    'struct<__index_level_0__:bigint,c0:bigint,c1:bigint>'
    >>> inferred.index_fields
    [InternalField(dtype=int64, struct_field=StructField('__index_level_0__', LongType(), True))]

    >>> def func() -> ps.DataFrame[pdf.index.dtype, pdf.dtypes]:
    ...     pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [dtype('int64'), dtype('int64'),
     CategoricalDtype(categories=[3, 4, 5], ordered=False, categories_dtype=int64)]
    >>> inferred.spark_type.simpleString()
    'struct<__index_level_0__:bigint,c0:bigint,c1:bigint>'
    >>> inferred.index_fields
    [InternalField(dtype=int64, struct_field=StructField('__index_level_0__', LongType(), True))]

    >>> def func() -> ps.DataFrame[
    ...     ("index", CategoricalDtype(categories=[3, 4, 5], ordered=False)),
    ...     [("id", int), ("A", int)]]:
    ...     pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [CategoricalDtype(categories=[3, 4, 5], ordered=False, categories_dtype=int64),
     dtype('int64'), dtype('int64')]
    >>> inferred.spark_type.simpleString()
    'struct<index:bigint,id:bigint,A:bigint>'
    >>> inferred.index_fields
    [InternalField(dtype=category, struct_field=StructField('index', LongType(), True))]

    >>> def func() -> ps.DataFrame[
    ...         (pdf.index.name, pdf.index.dtype), zip(pdf.columns, pdf.dtypes)]:
    ...     pass
    >>> inferred = infer_return_type(func)
    >>> inferred.dtypes
    [dtype('int64'), dtype('int64'),
     CategoricalDtype(categories=[3, 4, 5], ordered=False, categories_dtype=int64)]
    >>> inferred.spark_type.simpleString()
    'struct<__index_level_0__:bigint,a:bigint,b:bigint>'
    >>> inferred.index_fields
    [InternalField(dtype=int64, struct_field=StructField('__index_level_0__', LongType(), True))]
    """
    # We should re-import to make sure the class 'SeriesType' is not treated as a class
    # within this module locally. See Series.__class_getitem__ which imports this class
    # canonically.
    from pyspark.pandas.internal import InternalField, SPARK_INDEX_NAME_FORMAT
    from pyspark.pandas.typedef import SeriesType, NameTypeHolder, IndexNameTypeHolder
    from pyspark.pandas.utils import name_like_string

    tpe = get_type_hints(f).get("return", None)

    if tpe is None:
        raise ValueError("A return value is required for the input function")

    if hasattr(tpe, "__origin__") and issubclass(tpe.__origin__, SeriesType):
        tpe = tpe.__args__[0]
        if isinstance(tpe, type) and issubclass(tpe, NameTypeHolder):
            tpe = tpe.tpe
        dtype, spark_type = pandas_on_spark_type(tpe)
        return SeriesType(dtype, spark_type)

    # Note that, DataFrame type hints will create a Tuple.
    # Tuple has _name but other types have __name__
    name = getattr(tpe, "_name", getattr(tpe, "__name__", None))
    # Check if the name is Tuple.
    if name == "Tuple":
        tuple_type = tpe
        parameters = getattr(tuple_type, "__args__")

        index_parameters = [
            p for p in parameters if isclass(p) and issubclass(p, IndexNameTypeHolder)
        ]
        data_parameters = [p for p in parameters if p not in index_parameters]
        assert len(data_parameters) > 0, "Type hints for data must not be empty."

        index_fields = []
        if len(index_parameters) >= 1:
            for level, index_parameter in enumerate(index_parameters):
                index_name = index_parameter.name
                index_dtype, index_spark_type = pandas_on_spark_type(index_parameter.tpe)
                index_fields.append(
                    InternalField(
                        dtype=index_dtype,
                        struct_field=types.StructField(
                            name=(
                                index_name
                                if index_name is not None
                                else SPARK_INDEX_NAME_FORMAT(level)
                            ),
                            dataType=index_spark_type,
                        ),
                    )
                )
        else:
            # No type hint for index.
            assert len(index_parameters) == 0

        data_dtypes, data_spark_types = zip(
            *(
                (
                    pandas_on_spark_type(p.tpe)
                    if isclass(p) and issubclass(p, NameTypeHolder)
                    else pandas_on_spark_type(p)
                )
                for p in data_parameters
            )
        )
        data_names = [
            p.name if isclass(p) and issubclass(p, NameTypeHolder) else None
            for p in data_parameters
        ]
        data_fields = []
        for i, (data_name, data_dtype, data_spark_type) in enumerate(
            zip(data_names, data_dtypes, data_spark_types)
        ):
            data_fields.append(
                InternalField(
                    dtype=data_dtype,
                    struct_field=types.StructField(
                        name=name_like_string(data_name) if data_name is not None else ("c%s" % i),
                        dataType=data_spark_type,
                    ),
                )
            )

        return DataFrameType(index_fields=index_fields, data_fields=data_fields)

    tpes = pandas_on_spark_type(tpe)
    if tpes is None:
        return UnknownType(tpe)
    else:
        return ScalarType(*tpes)


# TODO: once pandas exposes a typing module like numpy.typing, we should deprecate
#   this logic and migrate to it by implementing the typing module in pandas API on Spark.


def create_type_for_series_type(param: Any) -> Type[SeriesType]:
    """
    Supported syntax:

    >>> str(ps.Series[float]).endswith("SeriesType[float]")
    True
    """
    from pyspark.pandas.typedef import NameTypeHolder

    new_class: Type[NameTypeHolder]
    if isinstance(param, ExtensionDtype):
        new_class = type(NameTypeHolder.short_name, (NameTypeHolder,), {})
        new_class.tpe = param  # type: ignore[assignment]
    else:
        if LooseVersion(pd.__version__) < "3.0.0":
            new_class = param.type if isinstance(param, np.dtype) else param
        else:
            new_class = param

    return SeriesType[new_class]  # type: ignore[valid-type]


# TODO: Remove this variadic-generic hack by tuple once ww drop Python up to 3.9.
#   See also PEP 646. One problem is that pandas doesn't inherits Generic[T]
#   so we might have to leave this hack only for monkey-patching pandas DataFrame.
def create_tuple_for_frame_type(params: Any) -> object:
    """
    This is a workaround to support variadic generic in DataFrame.

    See https://github.com/python/typing/issues/193
    we always wraps the given type hints by a tuple to mimic the variadic generic.

    Supported syntax:

    >>> import pandas as pd
    >>> pdf = pd.DataFrame({'a': range(1)})

    Typing data columns only:

        >>> ps.DataFrame[float, float]  # doctest: +ELLIPSIS
        typing.Tuple[...NameType, ...NameType]
        >>> ps.DataFrame[pdf.dtypes]  # doctest: +ELLIPSIS
        typing.Tuple[...NameType]
        >>> ps.DataFrame["id": int, "A": int]  # doctest: +ELLIPSIS
        typing.Tuple[...NameType, ...NameType]
        >>> ps.DataFrame[zip(pdf.columns, pdf.dtypes)]  # doctest: +ELLIPSIS
        typing.Tuple[...NameType]

    Typing data columns with an index:

        >>> ps.DataFrame[int, [int, int]]  # doctest: +ELLIPSIS
        typing.Tuple[...IndexNameType, ...NameType, ...NameType]
        >>> ps.DataFrame[pdf.index.dtype, pdf.dtypes]  # doctest: +ELLIPSIS
        typing.Tuple[...IndexNameType, ...NameType]
        >>> ps.DataFrame[("index", int), [("id", int), ("A", int)]]  # doctest: +ELLIPSIS
        typing.Tuple[...IndexNameType, ...NameType, ...NameType]
        >>> ps.DataFrame[(pdf.index.name, pdf.index.dtype), zip(pdf.columns, pdf.dtypes)]
        ... # doctest: +ELLIPSIS
        typing.Tuple[...IndexNameType, ...NameType]

    Typing data columns with an Multi-index:
        >>> arrays = [[1, 1, 2], ['red', 'blue', 'red']]
        >>> idx = pd.MultiIndex.from_arrays(arrays, names=('number', 'color'))
        >>> pdf = pd.DataFrame({'a': range(3)}, index=idx)
        >>> ps.DataFrame[[int, int], [int, int]]  # doctest: +ELLIPSIS
        typing.Tuple[...IndexNameType, ...IndexNameType, ...NameType, ...NameType]
        >>> ps.DataFrame[pdf.index.dtypes, pdf.dtypes]  # doctest: +ELLIPSIS, +SKIP
        typing.Tuple[...IndexNameType, ...NameType]
        >>> ps.DataFrame[[("index-1", int), ("index-2", int)], [("id", int), ("A", int)]]
        ... # doctest: +ELLIPSIS
        typing.Tuple[...IndexNameType, ...IndexNameType, ...NameType, ...NameType]
        >>> ps.DataFrame[zip(pdf.index.names, pdf.index.dtypes), zip(pdf.columns, pdf.dtypes)]
        ... # doctest: +ELLIPSIS, +SKIP
        typing.Tuple[...IndexNameType, ...NameType]
    """
    return Tuple[_to_type_holders(params)]


def _to_type_holders(params: Any) -> Tuple:
    from pyspark.pandas.typedef import NameTypeHolder, IndexNameTypeHolder

    is_with_index = (
        isinstance(params, tuple)
        and len(params) == 2
        and isinstance(params[1], (zip, list, pd.Series))
    )

    if is_with_index:
        # With index
        #   DataFrame[index_type, [type, ...]]
        #   DataFrame[dtype instance, dtypes instance]
        #   DataFrame[[index_type, ...], [type, ...]]
        #   DataFrame[dtypes instance, dtypes instance]
        #   DataFrame[(index_name, index_type), [(name, type), ...]]
        #   DataFrame[(index_name, index_type), zip(names, types)]
        #   DataFrame[[(index_name, index_type), ...], [(name, type), ...]]
        #   DataFrame[zip(index_names, index_types), zip(names, types)]
        def is_list_of_pairs(p: Any) -> bool:
            return (
                isinstance(p, list)
                and len(p) >= 1
                and all(isinstance(param, tuple) and (len(param) == 2) for param in p)
            )

        index_params = params[0]
        if isinstance(index_params, tuple) and len(index_params) == 2:
            # DataFrame[("index", int), ...]
            index_params = [index_params]

        if is_list_of_pairs(index_params):
            # DataFrame[[("index", int), ("index-2", int)], ...]
            index_params = tuple(slice(name, tpe) for name, tpe in index_params)

        index_types = _new_type_holders(index_params, IndexNameTypeHolder)

        data_types = params[1]
        if is_list_of_pairs(data_types):
            # DataFrame[..., [("id", int), ("A", int)]]
            data_types = tuple(slice(*data_type) for data_type in data_types)

        data_types = _new_type_holders(data_types, NameTypeHolder)

        return index_types + data_types
    else:
        # Without index
        #   DataFrame[type, type, ...]
        #   DataFrame[name: type, name: type, ...]
        #   DataFrame[dtypes instance]
        #   DataFrame[zip(names, types)]
        return _new_type_holders(params, NameTypeHolder)


def _new_type_holders(
    params: Any, holder_clazz: Type[Union[NameTypeHolder, IndexNameTypeHolder]]
) -> Tuple:
    if isinstance(params, zip):
        #   DataFrame[zip(names, types)]
        params = tuple(slice(name, tpe) for name, tpe in params)

    if isinstance(params, Iterable):
        #   DataFrame[type, type, ...]
        #   DataFrame[name: type, name: type, ...]
        #   DataFrame[dtypes instance]
        params = tuple(params)
    else:
        #   DataFrame[type, type]
        #   DataFrame[name: type]
        params = (params,)

    is_named_params = all(
        isinstance(param, slice) and param.step is None and param.stop is not None
        for param in params
    )
    if sys.version_info < (3, 11):
        is_unnamed_params = all(
            not isinstance(param, slice) and not isinstance(param, Iterable) for param in params
        )
    else:
        # PEP 646 changes `GenericAlias` instances into iterable ones at Python 3.11
        is_unnamed_params = all(
            not isinstance(param, slice)
            and (
                not isinstance(param, Iterable)
                or isinstance(param, typing.GenericAlias)  # type: ignore[attr-defined]
                or isinstance(param, typing._GenericAlias)  # type: ignore[attr-defined]
            )
            for param in params
        )

    if is_named_params:
        # DataFrame["id": int, "A": int]
        new_params = []
        for param in params:
            new_param: Type[Union[NameTypeHolder, IndexNameTypeHolder]] = type(
                holder_clazz.short_name, (holder_clazz,), {}
            )
            new_param.name = param.start
            if LooseVersion(pd.__version__) < "3.0.0":
                if isinstance(param.stop, ExtensionDtype):
                    new_param.tpe = param.stop  # type: ignore[assignment]
                else:
                    # When the given argument is a numpy's dtype instance.
                    new_param.tpe = (
                        param.stop.type  # type: ignore[assignment]
                        if isinstance(param.stop, np.dtype)
                        else param.stop
                    )
            else:
                new_param.tpe = param.stop
            new_params.append(new_param)
        return tuple(new_params)
    elif is_unnamed_params:
        # DataFrame[float, float]
        new_types = []
        for param in params:
            new_type: Type[Union[NameTypeHolder, IndexNameTypeHolder]] = type(
                holder_clazz.short_name, (holder_clazz,), {}
            )
            if LooseVersion(pd.__version__) < "3.0.0":
                if isinstance(param, ExtensionDtype):
                    new_type.tpe = param  # type: ignore[assignment]
                else:
                    new_type.tpe = (
                        param.type  # type: ignore[assignment]
                        if isinstance(param, np.dtype)
                        else param
                    )
            else:
                new_type.tpe = param
            new_types.append(new_type)
        return tuple(new_types)
    else:
        raise TypeError(
            """Type hints should be specified as one of:
  - DataFrame[type, type, ...]
  - DataFrame[name: type, name: type, ...]
  - DataFrame[dtypes instance]
  - DataFrame[zip(names, types)]
  - DataFrame[index_type, [type, ...]]
  - DataFrame[(index_name, index_type), [(name, type), ...]]
  - DataFrame[dtype instance, dtypes instance]
  - DataFrame[(index_name, index_type), zip(names, types)]
  - DataFrame[[index_type, ...], [type, ...]]
  - DataFrame[[(index_name, index_type), ...], [(name, type), ...]]
  - DataFrame[dtypes instance, dtypes instance]
  - DataFrame[zip(index_names, index_types), zip(names, types)]\n"""
            + "However, got %s." % str(params)
        )


def _test() -> None:
    import doctest
    import sys
    import pyspark.pandas.typedef.typehints

    globs = pyspark.pandas.typedef.typehints.__dict__.copy()
    failure_count, test_count = doctest.testmod(
        pyspark.pandas.typedef.typehints,
        globs=globs,
        optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE,
    )
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pyspark.testing.pandasutils import PandasOnSparkTestCase
from pyspark.testing.sqlutils import SQLTestUtils
from pyspark.pandas.tests.groupby.test_split_apply import GroupbySplitApplyTestingFuncMixin


class GroupbySplitApplyStdMixin(GroupbySplitApplyTestingFuncMixin):
    def test_split_apply_combine_on_series(self):
        funcs = [
            ((False, False), ["std"]),
        ]
        self._test_split_apply_func(funcs)


class GroupbySplitApplyStdTests(
    GroupbySplitApplyStdMixin,
    PandasOnSparkTestCase,
    SQLTestUtils,
):
    pass


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Tests for PyArrow's pa.Array.cast() method using golden file comparison.

This test suite covers both safe=True (default) and safe=False modes:
- safe=True: Checks for overflows and unsafe conversions, raises errors on failure
- safe=False: Allows potentially unsafe conversions (truncation, overflow wrapping)

Each mode generates separate golden files to capture their different behaviors.

## Golden File Cell Format

Each cell in the golden file uses the value@type format:

- Success: [0, 1, null]@int16 - element values via scalar.as_py() and Arrow type after cast
- Failure: ERR@ArrowNotImplementedError - the exception class name

## Regenerating Golden Files

Set SPARK_GENERATE_GOLDEN_FILES=1 before running:

    SPARK_GENERATE_GOLDEN_FILES=1 python -m pytest \\
        python/pyspark/tests/upstream/pyarrow/test_pyarrow_array_cast.py

If package tabulate (https://pypi.org/project/tabulate/) is installed,
it will also regenerate the Markdown files.

## PyArrow Version Compatibility

The golden files capture behavior for a specific PyArrow version.
Regenerate when upgrading PyArrow, as cast support may change between versions.

Some known version-dependent behaviors:
| Feature                                 | PyArrow < 19   | PyArrow 19-20  | PyArrow >= 21  |
|-----------------------------------------|----------------|----------------|----------------|
| struct cast: field name mismatch        | ArrowTypeError | supported      | supported      |
| struct cast: field reorder              | ArrowTypeError | ArrowTypeError | supported      |
| float16 scalar.as_py()                  | np.float16     | np.float16     | Python float   |
| pa.array(floats, pa.float16()) natively | requires numpy | requires numpy | native         |
"""

import inspect
import os
import platform
import unittest
from decimal import Decimal
from typing import Callable, List, Optional

from pyspark.loose_version import LooseVersion
from pyspark.testing.utils import (
    have_pyarrow,
    have_pandas,
    have_numpy,
    pyarrow_requirement_message,
    pandas_requirement_message,
    numpy_requirement_message,
)
from pyspark.testing.goldenutils import GoldenFileTestMixin

if have_pyarrow:
    import pyarrow as pa
if have_numpy:
    import numpy as np


# ============================================================
# Base Test Class
# ============================================================


class _PyArrowCastTestBase(GoldenFileTestMixin, unittest.TestCase):
    """Base class for PyArrow cast golden file tests with shared helpers."""

    @staticmethod
    def _make_float16_array(values):
        """
        Create a float16 PyArrow array from Python float values.

        PyArrow < 21 requires numpy.float16 instances to create float16 arrays,
        while PyArrow >= 21 accepts Python floats directly.
        """
        if LooseVersion(pa.__version__) >= LooseVersion("21.0.0"):
            return pa.array(values, pa.float16())
        else:
            import numpy as np

            np_values = [np.float16(v) if v is not None else None for v in values]
            return pa.array(np_values, pa.float16())

    def _try_cast(self, src_arr, tgt_type, safe=True):
        """
        Try casting a source array to target type and return a value@type string.

        Uses repr_value() from GoldenFileTestMixin, which formats PyArrow arrays
        as "[val1, val2, null]@arrow_type" using each scalar's as_py() value.

        Parameters
        ----------
        src_arr : pa.Array
            Source PyArrow array to cast
        tgt_type : pa.DataType
            Target PyArrow type
        safe : bool, default True
            If True, check for overflows and unsafe conversions.
            If False, allow potentially unsafe conversions.

        Returns
        -------
        str
            On success: "[val1, val2, null]@arrow_type"
                e.g. "[0, 1, -1, 127, -128, null]@int16"
            On failure: "ERR@<exception_class_name>"
                e.g. "ERR@ArrowNotImplementedError"
        """
        try:
            result = src_arr.cast(tgt_type, safe=safe)
            return self.repr_value(result, max_len=0)
        except Exception as e:
            return f"ERR@{type(e).__name__}"

    def compare_or_generate_golden_matrix(
        self,
        row_names: List[str],
        col_names: List[str],
        compute_cell: Callable[[str, str], str],
        golden_file_prefix: str,
        index_name: str = "source \\ target",
        overrides: Optional[dict[tuple[str, str], str]] = None,
    ) -> None:
        """
        Run a matrix of computations and compare against (or generate) a golden file.

        1. If SPARK_GENERATE_GOLDEN_FILES=1, compute every cell, build a
           DataFrame, and save it as the new golden CSV / Markdown file.
        2. Otherwise, load the existing golden file and assert that every cell
           matches the freshly computed value.

        Parameters
        ----------
        row_names : list[str]
            Ordered row labels (becomes the DataFrame index).
        col_names : list[str]
            Ordered column labels.
        compute_cell : (row_name, col_name) -> str
            Function that computes the string result for one cell.
        golden_file_prefix : str
            Prefix for the golden CSV/MD files (without extension).
            Files are placed in the same directory as the concrete test file.
        index_name : str, default "source \\ target"
            Name for the index column in the golden file.
        overrides : dict[(row, col) -> str], optional
            Version-specific expected values that take precedence over the golden
            file.  Use this to document known behavioral differences across
            library versions (e.g. PyArrow 18 vs 22) directly in the test code,
            so that the same golden file works for multiple versions.
        """
        generating = self.is_generating_golden()

        test_dir = os.path.dirname(inspect.getfile(type(self)))
        golden_csv = os.path.join(test_dir, f"{golden_file_prefix}.csv")
        golden_md = os.path.join(test_dir, f"{golden_file_prefix}.md")

        golden = None
        if not generating:
            golden = self.load_golden_csv(golden_csv)

        errors = []
        results = {}

        for row_name in row_names:
            for col_name in col_names:
                result = compute_cell(row_name, col_name)
                results[(row_name, col_name)] = result

                if not generating:
                    if overrides and (row_name, col_name) in overrides:
                        expected = overrides[(row_name, col_name)]
                    else:
                        expected = golden.loc[row_name, col_name]
                    if expected != result:
                        errors.append(
                            f"{row_name} -> {col_name}: expected '{expected}', got '{result}'"
                        )

        if generating:
            import pandas as pd

            index = pd.Index(row_names, name=index_name)
            df = pd.DataFrame(index=index)
            for col_name in col_names:
                df[col_name] = [results[(row, col_name)] for row in row_names]
            self.save_golden(df, golden_csv, golden_md)
        else:
            self.assertEqual(
                len(errors),
                0,
                f"\n{len(errors)} golden file mismatches:\n" + "\n".join(errors),
            )


# ============================================================
# Scalar Type Cast Tests
# ============================================================


@unittest.skipIf(
    not have_pyarrow
    or not have_pandas
    or not have_numpy
    or LooseVersion(np.__version__) < LooseVersion("2.0.0"),
    pyarrow_requirement_message or pandas_requirement_message or numpy_requirement_message,
)
class PyArrowScalarTypeCastTests(_PyArrowCastTestBase):
    """
    Tests all scalar-to-scalar type cast combinations via golden file comparison.

    Covers:
    - Integers: int8, int16, int32, int64, uint8, uint16, uint32, uint64
    - Floats: float16, float32, float64
    - Boolean: bool
    - Strings: string, large_string
    - Binary: binary, large_binary, fixed_size_binary
    - Decimal: decimal128, decimal256
    - Date: date32, date64
    - Timestamp: timestamp(s/ms/us/ns), with/without timezone
    - Duration: duration(s/ms/us/ns)
    - Time: time32(s/ms), time64(us/ns)
    """

    # ----- source case helpers -----

    def _signed_int_cases(self, pa_type, max_val, min_val):
        name = self.repr_type(pa_type)
        return [
            (f"{name}:standard", pa.array([0, 1, None], pa_type)),
            (f"{name}:negative", pa.array([-1, None], pa_type)),
            (f"{name}:max_min", pa.array([max_val, min_val, None], pa_type)),
        ]

    def _unsigned_int_cases(self, pa_type, max_val):
        name = self.repr_type(pa_type)
        return [
            (f"{name}:standard", pa.array([0, 1, None], pa_type)),
            (f"{name}:max", pa.array([max_val, None], pa_type)),
        ]

    def _standard_negative_cases(self, pa_type):
        name = self.repr_type(pa_type)
        return [
            (f"{name}:standard", pa.array([0, 1, None], pa_type)),
            (f"{name}:negative", pa.array([-1, None], pa_type)),
        ]

    # ----- target types -----

    @staticmethod
    def _get_target_types():
        return [
            # Integers
            pa.int8(),
            pa.int16(),
            pa.int32(),
            pa.int64(),
            pa.uint8(),
            pa.uint16(),
            pa.uint32(),
            pa.uint64(),
            # Floats
            pa.float16(),
            pa.float32(),
            pa.float64(),
            # Boolean
            pa.bool_(),
            # Strings
            pa.string(),
            pa.large_string(),
            # Binary
            pa.binary(),
            pa.large_binary(),
            pa.binary(16),
            # Decimal
            pa.decimal128(38, 10),
            pa.decimal256(76, 10),
            # Date
            pa.date32(),
            pa.date64(),
            # Timestamp (NTZ)
            pa.timestamp("s"),
            pa.timestamp("ms"),
            pa.timestamp("us"),
            pa.timestamp("ns"),
            # Timestamp (UTC)
            pa.timestamp("s", tz="UTC"),
            pa.timestamp("ms", tz="UTC"),
            pa.timestamp("us", tz="UTC"),
            pa.timestamp("ns", tz="UTC"),
            # Timestamp (other TZ)
            pa.timestamp("s", tz="America/New_York"),
            pa.timestamp("s", tz="Asia/Shanghai"),
            # Duration
            pa.duration("s"),
            pa.duration("ms"),
            pa.duration("us"),
            pa.duration("ns"),
            # Time
            pa.time32("s"),
            pa.time32("ms"),
            pa.time64("us"),
            pa.time64("ns"),
        ]

    # ----- source arrays -----

    def _get_source_arrays(self):
        """
        Create test arrays for all scalar types, split into separate edge-case
        categories.

        Each source type has multiple test cases so that edge-case values don't
        mask the behavior of other values.  For example int8:standard tests
        [0, 1, None] and int8:negative tests [-1, None] separately, so you can
        see that int8:standard -> uint8 succeeds while int8:negative -> uint8
        fails.
        """
        cases = []

        # --- Signed integers: standard, negative, max_min ---
        cases += self._signed_int_cases(pa.int8(), 127, -128)
        cases += self._signed_int_cases(pa.int16(), 32767, -32768)
        cases += self._signed_int_cases(pa.int32(), 2147483647, -2147483648)
        cases += self._signed_int_cases(pa.int64(), 2147483647, -2147483648)

        # --- Unsigned integers: standard, max ---
        cases += self._unsigned_int_cases(pa.uint8(), 255)
        cases += self._unsigned_int_cases(pa.uint16(), 65535)
        cases += self._unsigned_int_cases(pa.uint32(), 4294967295)
        cases += self._unsigned_int_cases(pa.uint64(), 4294967295)

        # --- Floats: standard, special, fractional ---
        f16_name = self.repr_type(pa.float16())
        cases += [
            (f"{f16_name}:standard", self._make_float16_array([0.0, 1.5, -1.5, None])),
            (
                f"{f16_name}:special",
                self._make_float16_array([float("inf"), float("nan"), None]),
            ),
            (f"{f16_name}:fractional", self._make_float16_array([0.1, 0.9, None])),
        ]
        for ftype in [pa.float32(), pa.float64()]:
            fname = self.repr_type(ftype)
            cases += [
                (f"{fname}:standard", pa.array([0.0, 1.5, -1.5, None], ftype)),
                (
                    f"{fname}:special",
                    pa.array([float("inf"), float("-inf"), float("nan"), None], ftype),
                ),
                (f"{fname}:fractional", pa.array([0.1, 0.9, None], ftype)),
            ]

        # --- Boolean ---
        cases += [
            (f"{self.repr_type(pa.bool_())}:standard", pa.array([True, False, None], pa.bool_()))
        ]

        # --- Strings: numeric, alpha, unicode ---
        for stype in [pa.string(), pa.large_string()]:
            sname = self.repr_type(stype)
            cases += [
                (f"{sname}:numeric", pa.array(["0", "1", "-1", None], stype)),
                (f"{sname}:alpha", pa.array(["abc", "", None], stype)),
                (
                    f"{sname}:unicode",
                    pa.array(
                        ["\u4f60\u597d", "\u0645\u0631\u062d\u0628\u0627", "\U0001f389", None],
                        stype,
                    ),
                ),
            ]

        # --- Binary ---
        for btype in [pa.binary(), pa.large_binary()]:
            bname = self.repr_type(btype)
            cases += [
                (f"{bname}:standard", pa.array([b"\x00", b"\xff", b"hello", b"", None], btype)),
            ]
        fsb_type = pa.binary(16)
        cases += [
            (
                f"{self.repr_type(fsb_type)}:standard",
                pa.array([b"0123456789abcdef", b"\x00" * 16, None], fsb_type),
            ),
        ]

        # --- Decimal: standard, large ---
        for dtype in [pa.decimal128(38, 10), pa.decimal256(76, 10)]:
            dname = self.repr_type(dtype)
            cases += [
                (
                    f"{dname}:standard",
                    pa.array([Decimal("0"), Decimal("1.5"), Decimal("-1.5"), None], dtype),
                ),
                (f"{dname}:large", pa.array([Decimal("9999999999"), None], dtype)),
            ]

        # --- Date: standard, negative ---
        for dtype in [pa.date32(), pa.date64()]:
            dname = self.repr_type(dtype)
            if dtype == pa.date32():
                cases += [
                    (f"{dname}:standard", pa.array([0, 1, None], dtype)),
                    (f"{dname}:negative", pa.array([-1, None], dtype)),
                ]
            else:
                cases += [
                    (f"{dname}:standard", pa.array([0, 86400000, None], dtype)),
                    (f"{dname}:negative", pa.array([-86400000, None], dtype)),
                ]

        # --- Timestamps (all 10 variants): standard, negative ---
        ts_types = [
            pa.timestamp("s"),
            pa.timestamp("ms"),
            pa.timestamp("us"),
            pa.timestamp("ns"),
            pa.timestamp("s", tz="UTC"),
            pa.timestamp("ms", tz="UTC"),
            pa.timestamp("us", tz="UTC"),
            pa.timestamp("ns", tz="UTC"),
            pa.timestamp("s", tz="America/New_York"),
            pa.timestamp("s", tz="Asia/Shanghai"),
        ]
        for ttype in ts_types:
            cases += self._standard_negative_cases(ttype)

        # --- Duration: standard, negative ---
        for unit in ["s", "ms", "us", "ns"]:
            cases += self._standard_negative_cases(pa.duration(unit))

        # --- Time: standard, noon ---
        time_defs = [
            (pa.time32("s"), 1, 43200),
            (pa.time32("ms"), 1000, 43200000),
            (pa.time64("us"), 1000000, 43200000000),
            (pa.time64("ns"), 1000000000, 43200000000000),
        ]
        for ttype, one_unit, noon_val in time_defs:
            tname = self.repr_type(ttype)
            cases += [
                (f"{tname}:standard", pa.array([0, one_unit, None], ttype)),
                (f"{tname}:noon", pa.array([noon_val, None], ttype)),
            ]

        source_names = [name for name, _ in cases]
        source_arrays = dict(cases)
        return source_names, source_arrays

    # ----- overrides -----

    @classmethod
    def _overrides_safe(cls):
        """
        Build overrides for known version/platform-dependent behaviors (safe=True mode).

        PyArrow < 21: str(scalar) for float16 uses numpy's formatting
        (via np.float16), which may vary across numpy versions.
        The golden file uses PyArrow >= 21 output (Python float).
        We compute the expected values dynamically to handle this difference.
        """
        overrides = {}
        if LooseVersion(pa.__version__) < LooseVersion("21.0.0"):
            F16 = "float16"

            def f16_repr(values):
                arr = cls._make_float16_array(values)
                return cls.repr_value(arr, max_len=0)

            frac = f16_repr([0.1, 0.9, None])
            overrides.update(
                {
                    ("int16:max_min", F16): f16_repr([32767.0, -32768.0, None]),
                    ("float16:fractional", F16): frac,
                    ("float32:fractional", F16): frac,
                    ("float64:fractional", F16): frac,
                }
            )
        return overrides

    @classmethod
    def _overrides_unsafe(cls):
        """
        Build overrides for known PyArrow version/platform-dependent behaviors (safe=False).

        PyArrow < 21: str(scalar) for float16 uses numpy's formatting
        (via np.float16). Same dynamic computation approach as safe=True mode.

        ARM (aarch64/arm64): Unsafe float-to-integer casts produce different results
        than x86 due to IEEE 754 implementation-defined behavior:
        - ARM FCVT instructions saturate on overflow (inf->MAX, -inf->MIN, nan->0)
        - x86 SSE/AVX returns "integer indefinite" values
        - Negative float -> unsigned int: ARM saturates to 0, x86 may wrap
        The golden files are generated on x86; ARM values are hardcoded below.
        """
        overrides = {}
        if LooseVersion(pa.__version__) < LooseVersion("21.0.0"):
            F16 = "float16"

            def f16_repr(values):
                arr = cls._make_float16_array(values)
                return cls.repr_value(arr, max_len=0)

            frac = f16_repr([0.1, 0.9, None])
            overrides.update(
                {
                    ("int16:max_min", F16): f16_repr([32767.0, -32768.0, None]),
                    ("float16:fractional", F16): frac,
                    ("float32:fractional", F16): frac,
                    ("float64:fractional", F16): frac,
                }
            )

        if platform.machine() in ("aarch64", "arm64"):
            overrides.update(
                {
                    # float16:standard [0.0, 1.5, -1.5, None] -> unsigned int types
                    # -1.5 saturates to 0 on ARM, wraps on x86
                    ("float16:standard", "uint8"): "[0, 1, 0, None]@uint8",
                    ("float16:standard", "uint16"): "[0, 1, 0, None]@uint16",
                    ("float16:standard", "uint32"): "[0, 1, 0, None]@uint32",
                    ("float16:standard", "uint64"): "[0, 1, 0, None]@uint64",
                    # float16:special [inf, nan, None] -> integer types
                    ("float16:special", "int8"): "[-1, 0, None]@int8",
                    ("float16:special", "int16"): "[-1, 0, None]@int16",
                    ("float16:special", "int32"): "[2147483647, 0, None]@int32",
                    ("float16:special", "int64"): "[9223372036854775807, 0, None]@int64",
                    ("float16:special", "uint8"): "[255, 0, None]@uint8",
                    ("float16:special", "uint16"): "[65535, 0, None]@uint16",
                    ("float16:special", "uint32"): "[4294967295, 0, None]@uint32",
                    ("float16:special", "uint64"): "[18446744073709551615, 0, None]@uint64",
                    # float32:standard [0.0, 1.5, -1.5, None] -> unsigned int types
                    ("float32:standard", "uint8"): "[0, 1, 0, None]@uint8",
                    ("float32:standard", "uint32"): "[0, 1, 0, None]@uint32",
                    ("float32:standard", "uint64"): "[0, 1, 0, None]@uint64",
                    # float32:special [inf, -inf, nan, None] -> integer types
                    ("float32:special", "int8"): "[-1, 0, 0, None]@int8",
                    ("float32:special", "int16"): "[-1, 0, 0, None]@int16",
                    ("float32:special", "int32"): "[2147483647, -2147483648, 0, None]@int32",
                    (
                        "float32:special",
                        "int64",
                    ): "[9223372036854775807, -9223372036854775808, 0, None]@int64",
                    ("float32:special", "uint8"): "[255, 0, 0, None]@uint8",
                    ("float32:special", "uint16"): "[65535, 0, 0, None]@uint16",
                    ("float32:special", "uint32"): "[4294967295, 0, 0, None]@uint32",
                    ("float32:special", "uint64"): "[18446744073709551615, 0, 0, None]@uint64",
                    # float64:standard [0.0, 1.5, -1.5, None] -> unsigned int types
                    ("float64:standard", "uint8"): "[0, 1, 0, None]@uint8",
                    ("float64:standard", "uint16"): "[0, 1, 0, None]@uint16",
                    ("float64:standard", "uint64"): "[0, 1, 0, None]@uint64",
                    # float64:special [inf, -inf, nan, None] -> integer types
                    ("float64:special", "int8"): "[-1, 0, 0, None]@int8",
                    ("float64:special", "int16"): "[-1, 0, 0, None]@int16",
                    ("float64:special", "int32"): "[-1, 0, 0, None]@int32",
                    (
                        "float64:special",
                        "int64",
                    ): "[9223372036854775807, -9223372036854775808, 0, None]@int64",
                    ("float64:special", "uint8"): "[255, 0, 0, None]@uint8",
                    ("float64:special", "uint16"): "[65535, 0, 0, None]@uint16",
                    ("float64:special", "uint32"): "[4294967295, 0, 0, None]@uint32",
                    ("float64:special", "uint64"): "[18446744073709551615, 0, 0, None]@uint64",
                }
            )

            if platform.system() == "Darwin":
                # macOS ARM differs from Linux ARM for some unsafe casts due to
                # differences in LLVM code generation between the two platforms.
                overrides.update(
                    {
                        # negative float -> uint8/uint16: macOS ARM wraps (255/65535),
                        # Linux ARM saturates to 0
                        ("float16:standard", "uint8"): "[0, 1, 255, None]@uint8",
                        ("float16:standard", "uint16"): "[0, 1, 65535, None]@uint16",
                        ("float32:standard", "uint8"): "[0, 1, 255, None]@uint8",
                        ("float64:standard", "uint8"): "[0, 1, 255, None]@uint8",
                        ("float64:standard", "uint16"): "[0, 1, 65535, None]@uint16",
                        # negative float -> uint32: macOS ARM saturates to 0,
                        # Linux ARM wraps to 4294967295 (matching x86 golden)
                        ("float64:standard", "uint32"): "[0, 1, 0, None]@uint32",
                        # special float -> int32: macOS ARM saturates (INT32_MAX/MIN),
                        # Linux ARM gives -1/0
                        ("float64:special", "int32"): "[2147483647, -2147483648, 0, None]@int32",
                    }
                )

        return overrides

    # ----- test methods -----

    def test_scalar_cast_matrix(self):
        """Test all scalar-to-scalar type cast combinations with safe=True."""
        source_names, source_arrays = self._get_source_arrays()
        target_types = self._get_target_types()
        target_names = [self.repr_type(t) for t in target_types]
        target_lookup = dict(zip(target_names, target_types))

        self.compare_or_generate_golden_matrix(
            row_names=source_names,
            col_names=target_names,
            compute_cell=lambda src, tgt: self._try_cast(
                source_arrays[src], target_lookup[tgt], safe=True
            ),
            golden_file_prefix="golden_pyarrow_scalar_cast_safe",
            overrides=self._overrides_safe(),
        )

    def test_scalar_cast_matrix_unsafe(self):
        """Test all scalar-to-scalar type cast combinations with safe=False."""
        source_names, source_arrays = self._get_source_arrays()
        target_types = self._get_target_types()
        target_names = [self.repr_type(t) for t in target_types]
        target_lookup = dict(zip(target_names, target_types))

        self.compare_or_generate_golden_matrix(
            row_names=source_names,
            col_names=target_names,
            compute_cell=lambda src, tgt: self._try_cast(
                source_arrays[src], target_lookup[tgt], safe=False
            ),
            golden_file_prefix="golden_pyarrow_scalar_cast_unsafe",
            overrides=self._overrides_unsafe(),
        )


# ============================================================
# Nested Type Cast Tests
# ============================================================


@unittest.skipIf(
    not have_pyarrow
    or not have_pandas
    or not have_numpy
    or LooseVersion(np.__version__) < LooseVersion("2.0.0"),
    pyarrow_requirement_message or pandas_requirement_message or numpy_requirement_message,
)
class PyArrowNestedTypeCastTests(_PyArrowCastTestBase):
    """
    Tests nested/container type cast combinations via golden file comparison.

    Covers:
    - List variants: list, large_list, fixed_size_list
    - Map: map<key, value>
    - Struct: struct<fields...>
    - Container to scalar (should fail)
    """

    # ----- target types -----

    @staticmethod
    def _get_target_types():
        return [
            # List variants
            pa.list_(pa.int32()),
            pa.list_(pa.int64()),
            pa.list_(pa.string()),
            pa.large_list(pa.int32()),
            pa.large_list(pa.int64()),
            pa.list_(pa.int32(), 2),
            pa.list_(pa.int32(), 3),
            # Map
            pa.map_(pa.string(), pa.int32()),
            pa.map_(pa.string(), pa.int64()),
            # Struct variants: same names, type change, reorder, name mismatch
            pa.struct([("x", pa.int32()), ("y", pa.string())]),
            pa.struct([("x", pa.int64()), ("y", pa.string())]),
            pa.struct([("y", pa.string()), ("x", pa.int32())]),
            pa.struct([("a", pa.int32()), ("b", pa.string())]),
            # Scalar types (container -> scalar should fail)
            pa.string(),
            pa.int32(),
        ]

    # ----- source arrays -----

    def _get_source_arrays(self):
        """
        Create test arrays for nested/container types, split into edge-case
        categories.
        """
        list_int_type = pa.list_(pa.int32())
        struct_type = pa.struct([("x", pa.int32()), ("y", pa.string())])
        list_struct_type = pa.list_(struct_type)
        large_list_type = pa.large_list(pa.int32())
        fsl_type = pa.list_(pa.int32(), 2)
        map_type = pa.map_(pa.string(), pa.int32())

        rt = self.repr_type
        cases = [
            # --- list<int32> ---
            (f"{rt(list_int_type)}:standard", pa.array([[1, 2, 3], None], list_int_type)),
            (f"{rt(list_int_type)}:empty", pa.array([[], None], list_int_type)),
            (
                f"{rt(list_int_type)}:null_elem",
                pa.array([[None], [1, None, 3], None], list_int_type),
            ),
            # --- list<struct> ---
            (
                f"{rt(list_struct_type)}:standard",
                pa.array([[{"x": 1, "y": "a"}], None], list_struct_type),
            ),
            (
                f"{rt(list_struct_type)}:null_fields",
                pa.array([[{"x": None, "y": None}], [], None], list_struct_type),
            ),
            # --- large_list<int32> ---
            (f"{rt(large_list_type)}:standard", pa.array([[1, 2, 3], None], large_list_type)),
            (f"{rt(large_list_type)}:empty", pa.array([[], None], large_list_type)),
            (f"{rt(large_list_type)}:null_elem", pa.array([[None], None], large_list_type)),
            # --- fixed_size_list<int32, 2> ---
            (f"{rt(fsl_type)}:standard", pa.array([[1, 2], [3, 4], None], fsl_type)),
            (f"{rt(fsl_type)}:null_elem", pa.array([[None, None], None], fsl_type)),
            # --- map<string, int32> ---
            (f"{rt(map_type)}:standard", pa.array([[("a", 1), ("b", 2)], None], map_type)),
            (f"{rt(map_type)}:empty", pa.array([[], None], map_type)),
            # --- struct<x: int32, y: string> ---
            (f"{rt(struct_type)}:standard", pa.array([{"x": 1, "y": "a"}, None], struct_type)),
            (
                f"{rt(struct_type)}:null_fields",
                pa.array([{"x": None, "y": None}, None], struct_type),
            ),
        ]

        source_names = [name for name, _ in cases]
        source_arrays = dict(cases)
        return source_names, source_arrays

    # ----- overrides -----

    @staticmethod
    def _overrides_safe():
        """
        Build overrides for known PyArrow version-dependent behaviors (safe=True mode).

        PyArrow < 21: struct field reordering during cast is not supported,
            raises ArrowTypeError instead.
        PyArrow < 19: struct field name mismatch during cast is not supported,
            raises ArrowTypeError instead.
        """
        overrides = {}
        struct_sources = [
            "struct<x: int32, y: string>:standard",
            "struct<x: int32, y: string>:null_fields",
        ]
        # PyArrow < 21: struct field reorder not supported
        if LooseVersion(pa.__version__) < LooseVersion("21.0.0"):
            for src in struct_sources:
                overrides[(src, "struct<y: string, x: int32>")] = "ERR@ArrowTypeError"
        # PyArrow < 19: struct field name mismatch not supported
        if LooseVersion(pa.__version__) < LooseVersion("19.0.0"):
            for src in struct_sources:
                overrides[(src, "struct<a: int32, b: string>")] = "ERR@ArrowTypeError"
        return overrides

    @staticmethod
    def _overrides_unsafe():
        """
        Build overrides for known PyArrow version-dependent behaviors (safe=False mode).

        Same version-dependent struct behaviors apply in unsafe mode.
        Additional overrides may be needed for different PyArrow versions
        as safe=False behavior varies across versions.
        """
        overrides = {}
        struct_sources = [
            "struct<x: int32, y: string>:standard",
            "struct<x: int32, y: string>:null_fields",
        ]
        # PyArrow < 21: struct field reorder not supported
        if LooseVersion(pa.__version__) < LooseVersion("21.0.0"):
            for src in struct_sources:
                overrides[(src, "struct<y: string, x: int32>")] = "ERR@ArrowTypeError"
        # PyArrow < 19: struct field name mismatch not supported
        if LooseVersion(pa.__version__) < LooseVersion("19.0.0"):
            for src in struct_sources:
                overrides[(src, "struct<a: int32, b: string>")] = "ERR@ArrowTypeError"
        # Additional overrides will be discovered during cross-version testing
        return overrides

    # ----- test methods -----

    def test_nested_cast_matrix(self):
        """Test all nested type cast combinations with safe=True."""
        source_names, source_arrays = self._get_source_arrays()
        target_types = self._get_target_types()
        target_names = [self.repr_type(t) for t in target_types]
        target_lookup = dict(zip(target_names, target_types))

        self.compare_or_generate_golden_matrix(
            row_names=source_names,
            col_names=target_names,
            compute_cell=lambda src, tgt: self._try_cast(
                source_arrays[src], target_lookup[tgt], safe=True
            ),
            golden_file_prefix="golden_pyarrow_nested_cast_safe",
            overrides=self._overrides_safe(),
        )

    def test_nested_cast_matrix_unsafe(self):
        """Test all nested type cast combinations with safe=False."""
        source_names, source_arrays = self._get_source_arrays()
        target_types = self._get_target_types()
        target_names = [self.repr_type(t) for t in target_types]
        target_lookup = dict(zip(target_names, target_types))

        self.compare_or_generate_golden_matrix(
            row_names=source_names,
            col_names=target_names,
            compute_cell=lambda src, tgt: self._try_cast(
                source_arrays[src], target_lookup[tgt], safe=False
            ),
            golden_file_prefix="golden_pyarrow_nested_cast_unsafe",
            overrides=self._overrides_unsafe(),
        )


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import pandas as pd

from pyspark import pandas as ps
from pyspark.testing.pandasutils import PandasOnSparkTestCase


class GetDummiesObjectMixin:
    def test_get_dummies_object(self):
        pdf = pd.DataFrame(
            {
                "a": [1, 2, 3, 4, 4, 3, 2, 1],
                # 'a': pd.Categorical([1, 2, 3, 4, 4, 3, 2, 1]),
                "b": list("abcdabcd"),
                # 'c': pd.Categorical(list('abcdabcd')),
                "c": list("abcdabcd"),
            }
        )
        psdf = ps.from_pandas(pdf)

        # Explicitly exclude object columns
        self.assert_eq(
            ps.get_dummies(psdf, columns=["a", "c"]),
            pd.get_dummies(pdf, columns=["a", "c"], dtype=np.int8),
        )

        self.assert_eq(ps.get_dummies(psdf), pd.get_dummies(pdf, dtype=np.int8))
        self.assert_eq(ps.get_dummies(psdf.b), pd.get_dummies(pdf.b, dtype=np.int8))
        self.assert_eq(
            ps.get_dummies(psdf, columns=["b"]), pd.get_dummies(pdf, columns=["b"], dtype=np.int8)
        )

        self.assertRaises(KeyError, lambda: ps.get_dummies(psdf, columns=("a", "c")))
        self.assertRaises(TypeError, lambda: ps.get_dummies(psdf, columns="b"))

        # non-string names
        pdf = pd.DataFrame(
            {10: [1, 2, 3, 4, 4, 3, 2, 1], 20: list("abcdabcd"), 30: list("abcdabcd")}
        )
        psdf = ps.from_pandas(pdf)

        self.assert_eq(
            ps.get_dummies(psdf, columns=[10, 30]),
            pd.get_dummies(pdf, columns=[10, 30], dtype=np.int8),
        )

        self.assertRaises(TypeError, lambda: ps.get_dummies(psdf, columns=10))


class GetDummiesObjectTests(
    GetDummiesObjectMixin,
    PandasOnSparkTestCase,
):
    pass


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

from pyspark.sql.tests.pandas.streaming.test_pandas_transform_with_state import (
    TransformWithStateInPandasTestsMixin,
)
from pyspark import SparkConf
from pyspark.testing.connectutils import ReusedConnectTestCase


class TransformWithStateInPandasParityTests(
    TransformWithStateInPandasTestsMixin, ReusedConnectTestCase
):
    """
    Spark connect parity tests for TransformWithStateInPandas. Run every test case in
     `TransformWithStateInPandasTestsMixin` in spark connect mode.
    """

    @classmethod
    def conf(cls):
        # Due to multiple inheritance from the same level, we need to explicitly setting configs in
        # both TransformWithStateInPandasTestsMixin and ReusedConnectTestCase here
        cfg = SparkConf(loadDefaults=False)
        for base in cls.__bases__:
            if hasattr(base, "conf"):
                parent_cfg = base.conf()
                for k, v in parent_cfg.getAll():
                    cfg.set(k, v)

        # Extra removing config for connect suites
        if cfg._jconf is not None:
            cfg._jconf.remove("spark.master")

        return cfg

    @unittest.skip("Flaky in spark connect on CI. Skip for now. See SPARK-51368 for details.")
    def test_schema_evolution_scenarios(self):
        pass


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pandas as pd

from pyspark import pandas as ps
from pyspark.pandas.config import set_option, reset_option
from pyspark.testing.pandasutils import PandasOnSparkTestCase
from pyspark.testing.sqlutils import SQLTestUtils


class GroupBySACMixin:
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        set_option("compute.ops_on_diff_frames", True)

    @classmethod
    def tearDownClass(cls):
        reset_option("compute.ops_on_diff_frames")
        super().tearDownClass()

    def test_split_apply_combine_on_series(self):
        pdf1 = pd.DataFrame({"C": [0.362, 0.227, 1.267, -0.562], "B": [1, 2, 3, 4]})
        pdf2 = pd.DataFrame({"A": [1, 1, 2, 2]})
        psdf1 = ps.from_pandas(pdf1)
        psdf2 = ps.from_pandas(pdf2)

        for as_index in [True, False]:
            if as_index:

                def sort(df):
                    return df.sort_index()

            else:

                def sort(df):
                    return df.sort_values(list(df.columns)).reset_index(drop=True)

            with self.subTest(as_index=as_index):
                self.assert_eq(
                    sort(psdf1.groupby(psdf2.A, as_index=as_index).sum()),
                    sort(pdf1.groupby(pdf2.A, as_index=as_index).sum()),
                )
                self.assert_eq(
                    sort(psdf1.groupby(psdf2.A, as_index=as_index).B.sum()),
                    sort(pdf1.groupby(pdf2.A, as_index=as_index).B.sum()),
                )
                self.assert_eq(
                    sort(psdf1.groupby([psdf1.C, psdf2.A], as_index=as_index).sum()),
                    sort(pdf1.groupby([pdf1.C, pdf2.A], as_index=as_index).sum()),
                )
                self.assert_eq(
                    sort(psdf1.groupby([psdf1.C + 1, psdf2.A], as_index=as_index).sum()),
                    sort(pdf1.groupby([pdf1.C + 1, pdf2.A], as_index=as_index).sum()),
                )

        self.assert_eq(
            psdf1.B.groupby(psdf2.A).sum().sort_index(),
            pdf1.B.groupby(pdf2.A).sum().sort_index(),
        )
        self.assert_eq(
            (psdf1.B + 1).groupby(psdf2.A).sum().sort_index(),
            (pdf1.B + 1).groupby(pdf2.A).sum().sort_index(),
        )

        self.assert_eq(
            psdf1.B.groupby(psdf2.A.rename()).sum().sort_index(),
            pdf1.B.groupby(pdf2.A.rename()).sum().sort_index(),
        )
        self.assert_eq(
            psdf1.B.rename().groupby(psdf2.A).sum().sort_index(),
            pdf1.B.rename().groupby(pdf2.A).sum().sort_index(),
        )
        self.assert_eq(
            psdf1.B.rename().groupby(psdf2.A.rename()).sum().sort_index(),
            pdf1.B.rename().groupby(pdf2.A.rename()).sum().sort_index(),
        )


class GroupBySACTests(
    GroupBySACMixin,
    PandasOnSparkTestCase,
    SQLTestUtils,
):
    pass


if __name__ == "__main__":
    from pyspark.testing import main

    main()
